"""Test the DataRecoveryAgent"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import unittest
from collections import defaultdict
from mock import MagicMock as Mock, patch, ANY

from parameterized import parameterized, param

import DIRAC
from DIRAC import S_OK, S_ERROR, gLogger

from DIRAC.TransformationSystem.Agent.DataRecoveryAgent import DataRecoveryAgent
from DIRAC.TransformationSystem.Utilities.JobInfo import TaskInfoException

from DIRAC.tests.Utilities.utils import MatchStringWith

__RCSID__ = "$Id$"

MODULE_NAME = "DIRAC.TransformationSystem.Agent.DataRecoveryAgent"


class TestDRA(unittest.TestCase):
    """Test the DataRecoveryAgent"""

    dra = None

    @patch("DIRAC.Core.Base.AgentModule.PathFinder", new=Mock())
    @patch("DIRAC.ConfigurationSystem.Client.PathFinder.getSystemInstance", new=Mock())
    @patch("%s.ReqClient" % MODULE_NAME, new=Mock())
    def setUp(self):
        self.dra = DataRecoveryAgent(agentName="ILCTransformationSystem/DataRecoveryAgent", loadName="TestDRA")
        self.dra.transNoInput = ["MCGeneration"]
        self.dra.transWithInput = ["MCSimulation", "MCReconstruction"]
        self.dra.transformationTypes = ["MCGeneration", "MCSimulation", "MCReconstruction"]

        self.dra.reqClient = Mock(name="reqMock", spec=DIRAC.RequestManagementSystem.Client.ReqClient.ReqClient)
        self.dra.tClient = Mock(
            name="transMock", spec=DIRAC.TransformationSystem.Client.TransformationClient.TransformationClient
        )
        self.dra.fcClient = Mock(name="fcMock", spec=DIRAC.Resources.Catalog.FileCatalogClient.FileCatalogClient)
        self.dra.jobMon = Mock(
            name="jobMonMock", spec=DIRAC.WorkloadManagementSystem.Client.JobMonitoringClient.JobMonitoringClient
        )
        self.dra.printEveryNJobs = 10
        self.dra.log = Mock(name="LogMock")
        self.dra.addressTo = "myself"
        self.dra.addressFrom = "me"

    def tearDown(self):
        pass

    def getTestMock(self, nameID=0, jobID=1234567):
        """create a JobInfo object with mocks"""
        from DIRAC.TransformationSystem.Utilities.JobInfo import JobInfo

        testJob = Mock(name="jobInfoMock_%s" % nameID, spec=JobInfo)
        testJob.jobID = jobID
        testJob.tType = "testType"
        testJob.otherTasks = []
        testJob.errorCounts = []
        testJob.status = "Done"
        testJob.transFileStatus = ["Assigned", "Assigned"]
        testJob.inputFileStatus = ["Exists", "Exists"]
        testJob.outputFiles = ["/my/stupid/file.lfn", "/my/stupid/file2.lfn"]
        testJob.outputFileStatus = ["Exists", "Exists"]
        testJob.inputFiles = ["inputfile.lfn", "inputfile2.lfn"]
        testJob.pendingRequest = False
        testJob.getTaskInfo = Mock()
        return testJob

    @patch("DIRAC.Core.Base.AgentModule.PathFinder", new=Mock())
    @patch("DIRAC.ConfigurationSystem.Client.PathFinder.getSystemInstance", new=Mock())
    @patch("%s.ReqClient" % MODULE_NAME, new=Mock())
    def test_init(self):
        """test for DataRecoveryAgent initialisation...................................................."""
        res = DataRecoveryAgent(agentName="ILCTransformationSystem/DataRecoveryAgent", loadName="TestDRA")
        self.assertIsInstance(res, DataRecoveryAgent)

    def test_beginExecution(self):
        """test for DataRecoveryAgent beginExecution...................................................."""
        theOps = Mock(name="OpsInstance")
        theOps.getValue.side_effect = [["MCGeneration"], ["MCReconstruction", "Merge"]]
        with patch("DIRAC.TransformationSystem.Agent.DataRecoveryAgent.Operations", return_value=theOps):
            res = self.dra.beginExecution()
        assert isinstance(self.dra.transformationTypes, list)
        assert set(["MCGeneration", "MCReconstruction", "Merge"]) == set(self.dra.transformationTypes)
        assert set(["MCGeneration"]) == set(self.dra.transNoInput)
        assert set(["MCReconstruction", "Merge"]) == set(self.dra.transWithInput)
        self.assertFalse(self.dra.enabled)
        self.assertTrue(res["OK"])

    def test_getEligibleTransformations_success(self):
        """test for DataRecoveryAgent getEligibleTransformations success................................"""
        transInfoDict = dict(
            TransformationID=1234,
            TransformationName="TestProd12",
            Type="TestProd",
            AuthorDN="/some/cert/owner",
            AuthorGroup="Test_Prod",
        )
        self.dra.tClient.getTransformations = Mock(return_value=S_OK([transInfoDict]))
        res = self.dra.getEligibleTransformations(status="Active", typeList=["TestProds"])
        self.assertTrue(res["OK"])
        self.assertIsInstance(res["Value"], dict)
        vals = res["Value"]
        self.assertIn("1234", vals)
        self.assertIsInstance(vals["1234"], dict)
        self.assertEqual(transInfoDict, vals["1234"])

    def test_getEligibleTransformations_failed(self):
        """test for DataRecoveryAgent getEligibleTransformations failure................................"""
        self.dra.tClient.getTransformations = Mock(return_value=S_ERROR("No can Do"))
        res = self.dra.getEligibleTransformations(status="Active", typeList=["TestProds"])
        self.assertFalse(res["OK"])
        self.assertEqual("No can Do", res["Message"])

    def test_treatTransformation1(self):
        """test for DataRecoveryAgent treatTransformation success1.........................................."""
        getJobMock = Mock(name="getJobMOck")
        getJobMock.getJobs.return_value = (Mock(name="jobsMOck"), 50, 50)
        tinfoMock = Mock(name="infoMock", return_value=getJobMock)
        self.dra.checkAllJobs = Mock()
        # catch the printout to check path taken
        transInfoDict = dict(
            TransformationID=1234,
            TransformationName="TestProd12",
            Type="TestProd",
            AuthorDN="/some/cert/owner",
            AuthorGroup="Test_Prod",
        )
        with patch("%s.TransformationInfo" % MODULE_NAME, new=tinfoMock):
            self.dra.treatTransformation(1234, transInfoDict)  # returns None
        # check we start with the summary right away
        for _name, args, _kwargs in self.dra.log.notice.mock_calls:
            self.assertNotIn("Getting Tasks:", str(args))

    def test_treatTransformation2(self):
        """test for DataRecoveryAgent treatTransformation success2.........................................."""
        getJobMock = Mock(name="getJobMOck")
        getJobMock.getJobs.return_value = (Mock(name="jobsMock"), 50, 50)
        tinfoMock = Mock(name="infoMock", return_value=getJobMock)
        self.dra.checkAllJobs = Mock()
        # catch the printout to check path taken
        transInfoDict = dict(
            TransformationID=1234,
            TransformationName="TestProd12",
            Type="MCSimulation",
            AuthorDN="/some/cert/owner",
            AuthorGroup="Test_Prod",
        )
        with patch("%s.TransformationInfo" % MODULE_NAME, new=tinfoMock):
            self.dra.treatTransformation(1234, transInfoDict)  # returns None
        self.dra.log.notice.assert_any_call(MatchStringWith("Getting tasks..."))

    def test_treatTransformation3(self):
        """test for DataRecoveryAgent treatTransformation skip.............................................."""
        getJobMock = Mock(name="getJobMOck")
        getJobMock.getJobs.return_value = (Mock(name="jobsMock"), 50, 50)
        self.dra.checkAllJobs = Mock()
        self.dra.jobCache[1234] = (50, 50)
        # catch the printout to check path taken
        transInfoDict = dict(
            TransformationID=1234,
            TransformationName="TestProd12",
            Type="TestProd",
            AuthorDN="/some/cert/owner",
            AuthorGroup="Test_Prod",
        )

        with patch("%s.TransformationInfo" % MODULE_NAME, autospec=True, return_value=getJobMock):
            self.dra.treatTransformation(transID=1234, transInfoDict=transInfoDict)  # returns None
        self.dra.log.notice.assert_called_with(MatchStringWith("Skipping transformation 1234"))

    def test_checkJob(self):
        """test for DataRecoveryAgent checkJob No inputFiles............................................."""

        from DIRAC.TransformationSystem.Utilities.TransformationInfo import TransformationInfo

        tInfoMock = Mock(name="tInfoMock", spec=TransformationInfo)

        from DIRAC.TransformationSystem.Utilities.JobInfo import JobInfo

        # Test First option for MCGeneration
        tInfoMock.reset_mock()
        testJob = JobInfo(jobID=1234567, status="Failed", tID=123, tType="MCGeneration")
        testJob.outputFiles = ["/my/stupid/file.lfn"]
        testJob.outputFileStatus = ["Exists"]

        self.dra.checkJob(testJob, tInfoMock)
        self.assertIn("setJobDone", tInfoMock.method_calls[0])
        self.assertEqual(self.dra.todo["NoInputFiles"][0]["Counter"], 1)
        self.assertEqual(self.dra.todo["NoInputFiles"][1]["Counter"], 0)

        # Test Second option for MCGeneration
        tInfoMock.reset_mock()
        testJob.status = "Done"
        testJob.outputFileStatus = ["Missing"]
        self.dra.checkJob(testJob, tInfoMock)
        self.assertIn("setJobFailed", tInfoMock.method_calls[0])
        self.assertEqual(self.dra.todo["NoInputFiles"][0]["Counter"], 1)
        self.assertEqual(self.dra.todo["NoInputFiles"][1]["Counter"], 1)

        # Test Second option for MCGeneration
        tInfoMock.reset_mock()
        testJob.status = "Done"
        testJob.outputFileStatus = ["Exists"]
        self.dra.checkJob(testJob, tInfoMock)
        self.assertEqual(tInfoMock.method_calls, [])
        self.assertEqual(self.dra.todo["NoInputFiles"][0]["Counter"], 1)
        self.assertEqual(self.dra.todo["NoInputFiles"][1]["Counter"], 1)

    @parameterized.expand(
        [
            param(
                0,
                ["setJobDone", "setInputProcessed"],
                jStat="Failed",
                ifStat=["Exists"],
                ofStat=["Exists"],
                tFiStat=["Assigned"],
                others=True,
            ),
            param(
                1,
                ["setJobFailed"],
                ifStat=["Exists"],
                ofStat=["Missing"],
                others=True,
                ifProcessed=["/my/inputfile.lfn"],
            ),
            param(
                2, ["setJobFailed", "cleanOutputs"], ifStat=["Exists"], others=True, ifProcessed=["/my/inputfile.lfn"]
            ),
            param(3, ["cleanOutputs", "setJobFailed", "setInputDeleted"], ifStat=["Missing"]),
            param(4, ["cleanOutputs", "setJobFailed"], tFiStat=["Deleted"], ifStat=["Missing"]),
            param(5, ["setJobDone", "setInputProcessed"], jStat="Failed", ifStat=["Exists"], tFiStat=["Assigned"]),
            param(6, ["setJobDone"], jStat="Failed", ifStat=["Exists"], tFiStat=["Processed"]),
            param(7, ["setInputProcessed"], jStat="Done", ifStat=["Exists"], tFiStat=["Assigned"]),
            param(
                8,
                ["setInputMaxReset"],
                jStat="Failed",
                ifStat=["Exists"],
                ofStat=["Missing"],
                tFiStat=["Assigned"],
                errorCount=[14],
            ),
            param(
                9,
                ["setInputUnused"],
                jStat="Failed",
                ifStat=["Exists"],
                ofStat=["Missing"],
                tFiStat=["Assigned"],
                errorCount=[2],
            ),
            param(
                10,
                ["setInputUnused", "setJobFailed"],
                jStat="Done",
                ifStat=["Exists"],
                ofStat=["Missing"],
                tFiStat=["Assigned"],
            ),
            param(
                11,
                ["cleanOutputs", "setInputUnused"],
                jStat="Failed",
                ifStat=["Exists"],
                ofStat=["Missing", "Exists"],
                tFiStat=["Assigned"],
            ),
            param(
                12,
                ["cleanOutputs", "setInputUnused", "setJobFailed"],
                jStat="Done",
                ifStat=["Exists"],
                ofStat=["Missing", "Exists"],
                tFiStat=["Assigned"],
            ),
            param(
                13, ["setJobFailed"], jStat="Done", ifStat=["Exists"], ofStat=["Missing", "Missing"], tFiStat=["Unused"]
            ),
            param(14, [], jStat="Strange", ifStat=["Exists"], ofStat=["Exists"], tFiStat=["Processed"]),
            param(
                -1,
                [],
                jStat="Failed",
                ifStat=["Exists"],
                ofStat=["Missing", "Missing"],
                outFiles=["/my/stupid/file.lfn", "/my/stupid/file2.lfn"],
                tFiStat=["Processed"],
                others=True,
            ),
        ]
    )
    def test_checkJob_others_(
        self,
        counter,
        infoCalls,
        jID=1234567,
        jStat="Done",
        others=False,
        inFiles=["/my/inputfile.lfn"],
        outFiles=["/my/stupid/file.lfn"],
        ifStat=[],
        ofStat=["Exists"],
        ifProcessed=[],
        tFiStat=["Processed"],
        errorCount=[],
    ):
        from DIRAC.TransformationSystem.Utilities.TransformationInfo import TransformationInfo
        from DIRAC.TransformationSystem.Utilities.JobInfo import JobInfo

        tInfoMock = Mock(name="tInfoMock", spec=TransformationInfo)
        testJob = JobInfo(jobID=jID, status=jStat, tID=123, tType="MCSimulation")
        testJob.outputFiles = outFiles
        testJob.outputFileStatus = ofStat
        testJob.otherTasks = others
        testJob.inputFiles = inFiles
        testJob.inputFileStatus = ifStat
        testJob.transFileStatus = tFiStat
        testJob.errorCounts = errorCount
        self.dra.inputFilesProcessed = set(ifProcessed)
        self.dra.checkJob(testJob, tInfoMock)
        gLogger.notice("Testing counter", counter)
        gLogger.notice("Expecting calls", infoCalls)
        gLogger.notice("Called", tInfoMock.method_calls)
        assert len(infoCalls) == len(tInfoMock.method_calls)
        for index, infoCall in enumerate(infoCalls):
            self.assertIn(infoCall, tInfoMock.method_calls[index])
        for count in range(15):
            gLogger.notice("Checking Counter:", count)
            if count == counter:
                self.assertEqual(self.dra.todo["InputFiles"][count]["Counter"], 1)
            else:
                self.assertEqual(self.dra.todo["InputFiles"][count]["Counter"], 0)
        if 0 <= counter <= 2:
            assert set(testJob.inputFiles).issubset(self.dra.inputFilesProcessed)

    @parameterized.expand(
        [
            param(["cleanOutputs", "setJobFailed"]),
            param([], jID=667, jStat="Failed", ofStat=["Missing"]),
            param([], jID=668, jStat="Failed", ofStat=["Missing"], inFiles=["some"]),
        ]
    )
    def test_failHard(self, infoCalls, jID=666, jStat="Done", inFiles=None, ofStat=["Exists"]):
        """Test the job.failHard function."""
        from DIRAC.TransformationSystem.Utilities.TransformationInfo import TransformationInfo
        from DIRAC.TransformationSystem.Utilities.JobInfo import JobInfo

        tInfoMock = Mock(name="tInfoMock", spec=TransformationInfo)
        tInfoMock.reset_mock()
        testJob = JobInfo(jobID=666, status=jStat, tID=123, tType="MCSimulation")
        testJob.outputFiles = ["/my/stupid/file.lfn"]
        testJob.outputFileStatus = ofStat
        testJob.otherTasks = True
        testJob.inputFiles = inFiles
        testJob.inputFileExists = True
        testJob.fileStatus = "Processed"
        self.dra.inputFilesProcessed = set()
        self.dra._DataRecoveryAgent__failJobHard(testJob, tInfoMock)  # pylint: disable=protected-access, no-member
        gLogger.notice("Expecting calls", infoCalls)
        gLogger.notice("Called", tInfoMock.method_calls)
        assert len(infoCalls) == len(tInfoMock.method_calls)
        for index, infoCall in enumerate(infoCalls):
            self.assertIn(infoCall, tInfoMock.method_calls[index])
        if jStat == "Done":
            self.assertIn("Failing job %s" % jID, self.dra.notesToSend)
        else:
            self.assertNotIn("Failing job %s" % jID, self.dra.notesToSend)

    def test_notOnlyKeepers(self):
        """test for __notOnlyKeepers function"""

        funcToTest = self.dra._DataRecoveryAgent__notOnlyKeepers  # pylint: disable=protected-access, no-member
        self.assertTrue(funcToTest("MCGeneration"))

        self.dra.todo["InputFiles"][0]["Counter"] = 3  # keepers
        self.dra.todo["InputFiles"][3]["Counter"] = 0
        self.assertFalse(funcToTest("MCSimulation"))

        self.dra.todo["InputFiles"][0]["Counter"] = 3  # keepers
        self.dra.todo["InputFiles"][3]["Counter"] = 3
        self.assertTrue(funcToTest("MCSimulation"))

    def test_checkAllJob(self):
        """test for DataRecoveryAgent checkAllJobs ....................................................."""
        from DIRAC.TransformationSystem.Utilities.JobInfo import JobInfo

        # test with additional task dicts
        from DIRAC.TransformationSystem.Utilities.TransformationInfo import TransformationInfo

        tInfoMock = Mock(name="tInfoMock", spec=TransformationInfo)
        mockJobs = dict([(i, self.getTestMock()) for i in range(11)])
        mockJobs[2].pendingRequest = True
        mockJobs[3].getJobInformation = Mock(side_effect=(RuntimeError("ARGJob1"), None))
        mockJobs[4].getTaskInfo = Mock(side_effect=(TaskInfoException("ARG1"), None))
        taskDict = True
        lfnTaskDict = True
        self.dra.checkAllJobs(mockJobs, tInfoMock, taskDict, lfnTaskDict)
        self.dra.log.error.assert_any_call(MatchStringWith("+++++ Exception"), "ARGJob1")
        self.dra.log.error.assert_any_call(MatchStringWith("Skip Task, due to TaskInfoException: ARG1"))
        self.dra.log.reset_mock()

        # test inputFile None
        mockJobs = dict([(i, self.getTestMock(nameID=i)) for i in range(5)])
        mockJobs[1].inputFiles = []
        mockJobs[1].getTaskInfo = Mock(side_effect=(TaskInfoException("NoInputFile"), None))
        mockJobs[1].tType = "MCSimulation"
        tInfoMock.reset_mock()
        self.dra.checkAllJobs(mockJobs, tInfoMock, taskDict, lfnTaskDict=True)
        self.dra.log.notice.assert_any_call(MatchStringWith("Failing job hard"))

    def test_checkAllJob_2(self):
        """Test where failJobHard fails (via cleanOutputs)."""
        from DIRAC.TransformationSystem.Utilities.TransformationInfo import TransformationInfo

        tInfoMock = Mock(name="tInfoMock", spec=TransformationInfo)
        mockJobs = dict([(i, self.getTestMock()) for i in range(5)])
        mockJobs[2].pendingRequest = True
        mockJobs[3].getTaskInfo = Mock(side_effect=(TaskInfoException("ARGJob3"), None))
        mockJobs[3].inputFiles = []
        mockJobs[3].tType = "MCReconstruction"
        self.dra._DataRecoveryAgent__failJobHard = Mock(side_effect=(RuntimeError("ARGJob4"), None), name="FJH")
        self.dra.checkAllJobs(mockJobs, tInfoMock, tasksDict=True, lfnTaskDict=True)
        mockJobs[3].getTaskInfo.assert_called()
        self.dra._DataRecoveryAgent__failJobHard.assert_called()
        self.dra.log.error.assert_any_call(MatchStringWith("+++++ Exception"), "ARGJob4")
        self.dra.log.reset_mock()

    def test_execute(self):
        """test for DataRecoveryAgent execute .........................................................."""
        self.dra.treatTransformation = Mock()

        self.dra.transformationsToIgnore = [123, 456, 789]
        self.dra.jobCache = defaultdict(lambda: (0, 0))
        self.dra.jobCache[123] = (10, 10)
        self.dra.jobCache[124] = (10, 10)
        self.dra.jobCache[125] = (10, 10)

        # Eligible fails
        self.dra.log.reset_mock()
        self.dra.getEligibleTransformations = Mock(return_value=S_ERROR("outcast"))
        res = self.dra.execute()
        self.assertFalse(res["OK"])
        self.dra.log.error.assert_any_call(ANY, MatchStringWith("outcast"))
        self.assertEqual("Failure to get transformations", res["Message"])

        d123 = dict(
            TransformationID=123,
            TransformationName="TestProd123",
            Type="MCGeneration",
            AuthorDN="/some/cert/owner",
            AuthorGroup="Test_Prod",
        )
        d124 = dict(
            TransformationID=124,
            TransformationName="TestProd124",
            Type="MCGeneration",
            AuthorDN="/some/cert/owner",
            AuthorGroup="Test_Prod",
        )
        d125 = dict(
            TransformationID=125,
            TransformationName="TestProd125",
            Type="MCGeneration",
            AuthorDN="/some/cert/owner",
            AuthorGroup="Test_Prod",
        )

        # Eligible succeeds
        self.dra.log.reset_mock()
        self.dra.getEligibleTransformations = Mock(return_value=S_OK({123: d123, 124: d124, 125: d125}))
        res = self.dra.execute()
        self.assertTrue(res["OK"])
        self.dra.log.notice.assert_any_call(
            MatchStringWith("Will ignore the following transformations: [123, 456, 789]")
        )
        self.dra.log.notice.assert_any_call(MatchStringWith("Ignoring Transformation: 123"))
        self.dra.log.notice.assert_any_call(MatchStringWith("Running over Transformation: 124"))

        # Notes To Send
        self.dra.log.reset_mock()
        self.dra.getEligibleTransformations = Mock(return_value=S_OK({123: d123, 124: d124, 125: d125}))
        self.dra.notesToSend = "Da hast du deine Karte"
        sendmailMock = Mock()
        sendmailMock.sendMail.return_value = S_OK("Nice Card")
        notificationMock = Mock(return_value=sendmailMock)
        with patch("%s.NotificationClient" % MODULE_NAME, new=notificationMock):
            res = self.dra.execute()
        self.assertTrue(res["OK"])
        self.dra.log.notice.assert_any_call(
            MatchStringWith("Will ignore the following transformations: [123, 456, 789]")
        )
        self.dra.log.notice.assert_any_call(MatchStringWith("Ignoring Transformation: 123"))
        self.dra.log.notice.assert_any_call(MatchStringWith("Running over Transformation: 124"))
        self.assertNotIn(124, self.dra.jobCache)  # was popped
        self.assertIn(125, self.dra.jobCache)  # was not popped
        gLogger.notice("JobCache: %s" % self.dra.jobCache)

        # sending notes fails
        self.dra.log.reset_mock()
        self.dra.notesToSend = "Da hast du deine Karte"
        sendmailMock = Mock()
        sendmailMock.sendMail.return_value = S_ERROR("No stamp")
        notificationMock = Mock(return_value=sendmailMock)
        with patch("%s.NotificationClient" % MODULE_NAME, new=notificationMock):
            res = self.dra.execute()
        self.assertTrue(res["OK"])
        self.assertNotIn(124, self.dra.jobCache)  # was popped
        self.assertIn(125, self.dra.jobCache)  # was not popped
        self.dra.log.error.assert_any_call(MatchStringWith("Cannot send notification mail"), ANY)

        self.assertEqual("", self.dra.notesToSend)

    def test_printSummary(self):
        """test DataRecoveryAgent printSummary.........................................................."""
        self.dra.notesToSend = ""
        self.dra.printSummary()
        self.assertNotIn(" Other Tasks --> Keep                                    :     0", self.dra.notesToSend)

        self.dra.notesToSend = "Note This"
        self.dra.printSummary()

    def test_setPendingRequests_1(self):
        """Check the setPendingRequests function."""
        mockJobs = dict((i, self.getTestMock(jobID=i)) for i in range(11))
        reqMock = Mock()
        reqMock.Status = "Done"
        reqClient = Mock(name="reqMock", spec=DIRAC.RequestManagementSystem.Client.ReqClient.ReqClient)
        reqClient.readRequestsForJobs.return_value = S_OK({"Successful": {}})
        self.dra.reqClient = reqClient
        self.dra.setPendingRequests(mockJobs)
        for _index, mj in mockJobs.items():
            self.assertFalse(mj.pendingRequest)

    def test_setPendingRequests_2(self):
        """Check the setPendingRequests function."""
        mockJobs = dict((i, self.getTestMock(jobID=i)) for i in range(11))
        reqMock = Mock()
        reqMock.RequestID = 666
        reqClient = Mock(name="reqMock", spec=DIRAC.RequestManagementSystem.Client.ReqClient.ReqClient)
        reqClient.readRequestsForJobs.return_value = S_OK({"Successful": {6: reqMock}})
        reqClient.getRequestStatus.return_value = {"Value": "Done"}
        self.dra.reqClient = reqClient
        self.dra.setPendingRequests(mockJobs)
        for _index, mj in mockJobs.items():
            self.assertFalse(mj.pendingRequest)
        reqClient.getRequestStatus.assert_called_once_with(666)

    def test_setPendingRequests_3(self):
        """Check the setPendingRequests function."""
        mockJobs = dict((i, self.getTestMock(jobID=i)) for i in range(11))
        reqMock = Mock()
        reqMock.RequestID = 555
        reqClient = Mock(name="reqMock", spec=DIRAC.RequestManagementSystem.Client.ReqClient.ReqClient)
        reqClient.readRequestsForJobs.return_value = S_OK({"Successful": {5: reqMock}})
        reqClient.getRequestStatus.return_value = {"Value": "Pending"}
        self.dra.reqClient = reqClient
        self.dra.setPendingRequests(mockJobs)
        for index, mj in mockJobs.items():
            if index == 5:
                self.assertTrue(mj.pendingRequest)
            else:
                self.assertFalse(mj.pendingRequest)
        reqClient.getRequestStatus.assert_called_once_with(555)

    def test_setPendingRequests_Fail(self):
        """Check the setPendingRequests function."""
        mockJobs = dict((i, self.getTestMock(jobID=i)) for i in range(11))
        reqMock = Mock()
        reqMock.Status = "Done"
        reqClient = Mock(name="reqMock", spec=DIRAC.RequestManagementSystem.Client.ReqClient.ReqClient)
        reqClient.readRequestsForJobs.side_effect = (S_ERROR("Failure"), S_OK({"Successful": {}}))
        self.dra.reqClient = reqClient
        self.dra.setPendingRequests(mockJobs)
        for _index, mj in mockJobs.items():
            self.assertFalse(mj.pendingRequest)

    def test_getLFNStatus(self):
        """Check the getLFNStatus function."""
        mockJobs = dict((i, self.getTestMock(jobID=i)) for i in range(11))
        self.dra.fcClient.exists.return_value = S_OK(
            {"Successful": {"/my/stupid/file.lfn": True, "/my/stupid/file2.lfn": True}}
        )
        lfnExistence = self.dra.getLFNStatus(mockJobs)
        self.assertEqual(lfnExistence, {"/my/stupid/file.lfn": True, "/my/stupid/file2.lfn": True})

        self.dra.fcClient.exists.side_effect = (
            S_ERROR("args"),
            S_OK({"Successful": {"/my/stupid/file.lfn": True, "/my/stupid/file2.lfn": True}}),
        )
        lfnExistence = self.dra.getLFNStatus(mockJobs)
        self.assertEqual(lfnExistence, {"/my/stupid/file.lfn": True, "/my/stupid/file2.lfn": True})



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import socket
import time
import os

try:
    import selectors
except ImportError:
    import selectors2 as selectors

from DIRAC.Core.DISET.private.Transports.BaseTransport import BaseTransport
from DIRAC.FrameworkSystem.Client.Logger import gLogger
from DIRAC.Core.Utilities.ReturnValues import S_ERROR, S_OK


class PlainTransport(BaseTransport):
    def initAsClient(self):
        timeout = None
        if "timeout" in self.extraArgsDict:
            timeout = self.extraArgsDict["timeout"]
        try:
            self.oSocket = socket.create_connection(self.stServerAddress, timeout)
        except socket.error as e:
            if e.args[0] != 115:
                return S_ERROR("Can't connect: %s" % str(e))
            # Connect in progress
            sel = selectors.DefaultSelector()
            sel.register(self.oSocket, selectors.EVENT_READ)
            if not sel.select(timeout=self.extraArgsDict["timeout"]):
                self.oSocket.close()
                return S_ERROR("Connection timeout")
            errno = self.oSocket.getsockopt(socket.SOL_SOCKET, socket.SO_ERROR)
            if errno != 0:
                return S_ERROR("Can't connect: %s" % str((errno, os.strerror(errno))))
        self.remoteAddress = self.oSocket.getpeername()
        return S_OK(self.oSocket)

    def initAsServer(self):
        if not self.serverMode():
            raise RuntimeError("Must be initialized as server mode")
        try:
            self.oSocket = socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
        except socket.error:
            # IPv6 is probably disabled on this node, try IPv4 only instead
            self.oSocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        if self.bAllowReuseAddress:
            self.oSocket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.oSocket.bind(self.stServerAddress)
        self.oSocket.listen(self.iListenQueueSize)
        return S_OK(self.oSocket)

    def close(self):
        gLogger.debug("Closing socket")
        try:
            self.oSocket.shutdown(socket.SHUT_RDWR)
        except Exception:
            pass
        self.oSocket.close()

    def setClientSocket(self, oSocket):
        if self.serverMode():
            raise RuntimeError("Mustbe initialized as client mode")
        self.oSocket = oSocket
        if "timeout" in self.extraArgsDict:
            self.oSocket.settimeout(self.extraArgsDict["timeout"])
        self.remoteAddress = self.oSocket.getpeername()

    def acceptConnection(self):
        # HACK: Was = PlainTransport( self )
        oClientTransport = PlainTransport(self.stServerAddress)
        oClientSocket, stClientAddress = self.oSocket.accept()
        oClientTransport.setClientSocket(oClientSocket)
        return S_OK(oClientTransport)

    def _read(self, bufSize=4096, skipReadyCheck=False):
        start = time.time()
        timeout = False
        if "timeout" in self.extraArgsDict:
            timeout = self.extraArgsDict["timeout"]
        while True:
            if timeout:
                if time.time() - start > timeout:
                    return S_ERROR("Socket read timeout exceeded")
            try:
                data = self.oSocket.recv(bufSize)
                return S_OK(data)
            except socket.error as e:
                if e.errno == 11:
                    time.sleep(0.001)
                else:
                    return S_ERROR("Exception while reading from peer: %s" % str(e))
            except Exception as e:
                return S_ERROR("Exception while reading from peer: %s" % str(e))

    def _write(self, buf):
        sentBytes = 0
        timeout = False
        if "timeout" in self.extraArgsDict:
            timeout = self.extraArgsDict["timeout"]
        if timeout:
            start = time.time()
        while sentBytes < len(buf):
            try:
                if timeout:
                    if time.time() - start > timeout:
                        return S_ERROR("Socket write timeout exceeded")
                sent = self.oSocket.send(buf[sentBytes:])
                if sent == 0:
                    return S_ERROR("Connection closed by peer")
                if sent > 0:
                    sentBytes += sent
            except socket.error as e:
                if e.errno == 11:
                    time.sleep(0.001)
                else:
                    return S_ERROR("Exception while sending to peer: %s" % str(e))
            except Exception as e:
                return S_ERROR("Error while sending: %s" % str(e))
        return S_OK(sentBytes)


def checkSanity(*args, **kwargs):
    return S_OK({})


def delegate(delegationRequest, kwargs):
    """
    Check delegate!
    """
    return S_OK()



""" PEP

  PEP ( Policy Enforcement Point ) is the front-end of the whole Policy System.
  Any interaction with it must go through the PEP to ensure a smooth flow.

  Firstly, it loads the PDP ( Policy Decision Point ) which actually is the
  module doing all dirty work ( finding policies, running them, merging their
  results, etc... ). Indeed, the PEP takes the output of the PDP for a given set
  of parameters ( decisionParams ) and enforces the actions that apply ( also
  determined by the PDP output ).

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.ResourceStatusSystem.PolicySystem.PDP import PDP
from DIRAC.ResourceStatusSystem.Utilities import Utils
from DIRAC.Core.Utilities.ObjectLoader import ObjectLoader


class PEP(object):
    """PEP ( Policy Enforcement Point )"""

    def __init__(self, clients=dict()):
        """Constructor

        examples:
          >>> pep = PEP()
          >>> pep1 = PEP( { 'ResourceStatusClient' : ResourceStatusClient() } )
          >>> pep2 = PEP( { 'ResourceStatusClient' : ResourceStatusClient(), 'ClientY' : None } )

        :Parameters:
          **clients** - [ None, `dict` ]
            dictionary with clients to be used in the commands issued by the policies.
            If not defined, the commands will import them. It is a measure to avoid
            opening the same connection every time a policy is evaluated.

        """

        self.clients = dict(clients)

        # Creating the client in the PEP is a convenience for the PDP, that uses internally the RSS clients

        res = ObjectLoader().loadObject("DIRAC.ResourceStatusSystem.Client.ResourceStatusClient")
        if not res["OK"]:
            self.log.error("Failed to load ResourceStatusClient class: %s" % res["Message"])
            raise ImportError(res["Message"])
        rsClass = res["Value"]

        res = ObjectLoader().loadObject("DIRAC.ResourceStatusSystem.Client.ResourceManagementClient")
        if not res["OK"]:
            self.log.error("Failed to load ResourceManagementClient class: %s" % res["Message"])
            raise ImportError(res["Message"])
        rmClass = res["Value"]

        res = ObjectLoader().loadObject("DIRAC.ResourceStatusSystem.Client.SiteStatus")
        if not res["OK"]:
            self.log.error("Failed to load SiteStatus class: %s" % res["Message"])
            raise ImportError(res["Message"])
        ssClass = res["Value"]

        if "ResourceStatusClient" not in clients:
            self.clients["ResourceStatusClient"] = rsClass()
        if "ResourceManagementClient" not in clients:
            self.clients["ResourceManagementClient"] = rmClass()
        if "SiteStatus" not in clients:
            self.clients["SiteStatus"] = ssClass()

        # Pass to the PDP the clients that are going to be used on the Commands
        self.pdp = PDP(self.clients)

        self.log = gLogger

    def enforce(self, decisionParams):
        """Given a dictionary with decisionParams, it is passed to the PDP, which
        will return ( in case there is a/are positive match/es ) a dictionary containing
        three key-pair values: the original decisionParams ( `decisionParams` ), all
        the policies evaluated ( `singlePolicyResults` ) and the computed final result
        ( `policyCombinedResult` ).

        To know more about decisionParams, please read PDP.setup where the decisionParams
        are sanitized.

        examples:
           >>> pep.enforce( { 'element' : 'Site', 'name' : 'MySite' } )
           >>> pep.enforce( { 'element' : 'Resource', 'name' : 'myce.domain.ch' } )

        :Parameters:
          **decisionParams** - `dict`
            dictionary with the parameters that will be used to match policies.

        """
        if not decisionParams:
            self.log.warn("No decision params...?")
            return S_OK()

        standardParamsDict = {
            "element": None,
            "name": None,
            "elementType": None,
            "statusType": None,
            "status": None,
            "reason": None,
            "tokenOwner": None,
            # Last parameter allows policies to be de-activated
            "active": "Active",
        }

        standardParamsDict.update(decisionParams)

        if standardParamsDict["element"] is not None:
            self.log = gLogger.getSubLogger("PEP/%s" % standardParamsDict["element"])
            if standardParamsDict["name"] is not None:
                self.log = gLogger.getSubLogger(
                    "PEP/%s/%s" % (standardParamsDict["element"], standardParamsDict["name"])
                )
                self.log.verbose(
                    "Enforce - statusType: %s, status: %s"
                    % (standardParamsDict["statusType"], standardParamsDict["status"])
                )
        decisionParams = dict(standardParamsDict)

        # Setup PDP with new parameters dictionary
        self.pdp.setup(decisionParams)

        # Run policies, get decision, get actions to apply
        resDecisions = self.pdp.takeDecision()
        if not resDecisions["OK"]:
            self.log.error("Something went wrong, not enforcing policies", "%s" % decisionParams)
            return resDecisions
        resDecisions = resDecisions["Value"]

        # We take from PDP the decision parameters used to find the policies
        decisionParams = resDecisions["decisionParams"]
        policyCombinedResult = resDecisions["policyCombinedResult"]
        singlePolicyResults = resDecisions["singlePolicyResults"]

        # We have run the actions and at this point, we are about to execute the actions.
        # One more final check before proceeding
        isNotUpdated = self.__isNotUpdated(decisionParams)
        if not isNotUpdated["OK"]:
            return isNotUpdated

        for policyActionName, policyActionType in policyCombinedResult["PolicyAction"]:

            try:
                actionMod = Utils.voimport("DIRAC.ResourceStatusSystem.PolicySystem.Actions.%s" % policyActionType)
            except ImportError:
                self.log.error("Error importing %s action" % policyActionType)
                continue

            try:
                action = getattr(actionMod, policyActionType)
            except AttributeError:
                self.log.error("Error importing %s action class" % policyActionType)
                continue

            actionObj = action(
                policyActionName, decisionParams, policyCombinedResult, singlePolicyResults, self.clients
            )

            self.log.debug((policyActionName, policyActionType))

            actionResult = actionObj.run()
            if not actionResult["OK"]:
                self.log.error(actionResult["Message"])

        return S_OK(resDecisions)

    def __isNotUpdated(self, decisionParams):
        """Checks for the existence of the element as it was passed to the PEP. It may
        happen that while being the element processed by the PEP an user through the
        web interface or the CLI has updated the status for this particular element. As
        a result, the PEP would overwrite whatever the user had set. This check is not
        perfect, as still an user action can happen while executing the actions, but
        the probability is close to 0. However, if there is an action that takes seconds
        to be executed, this must be re-evaluated. !

        :Parameters:
          **decisionParams** - `dict`
            dictionary with the parameters that will be used to match policies

        :return: S_OK / S_ERROR

        """

        # Copy original dictionary and get rid of one key we cannot pass as kwarg
        selectParams = dict(decisionParams)
        del selectParams["element"]
        del selectParams["active"]

        # We expect to have an exact match. If not, then something has changed and
        # we cannot proceed with the actions.
        if decisionParams["element"] == "Site":
            unchangedRow = self.clients["SiteStatus"].getSiteStatuses([decisionParams["name"]])
        else:
            unchangedRow = self.clients["ResourceStatusClient"].selectStatusElement(
                decisionParams["element"], "Status", **selectParams
            )
        if not unchangedRow["OK"]:
            return unchangedRow

        if not unchangedRow["Value"]:
            msg = "%(name)s  ( %(status)s / %(statusType)s ) has been updated after PEP started running" % selectParams
            self.log.error(msg)
            return S_ERROR(msg)

        return S_OK()


# EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF



"""A reporter class to prepare reports and network accounting plots.

Supports:

* packet loss rate (standard and magnified),
* one-way delay, jitter, jitter over one-way delay

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import numpy as np

from DIRAC import S_OK
from DIRAC.AccountingSystem.Client.Types.Network import Network
from DIRAC.AccountingSystem.private.Plotters.BaseReporter import BaseReporter


class NetworkPlotter(BaseReporter):

    _typeName = "Network"
    _typeKeyFields = [dF[0] for dF in Network().definitionKeyFields]

    _reportPacketLossRateName = "Packet loss rate"

    def _reportPacketLossRate(self, reportRequest):

        selectFields = (
            self._getSelectStringForGrouping(reportRequest["groupingFields"]) + ", %s, %s, 100 - SUM(%s)/SUM(%s), 100",
            reportRequest["groupingFields"][1] + ["startTime", "bucketLength", "PacketLossRate", "entriesInBucket"],
        )
        retVal = self._getTimedData(
            reportRequest["startTime"],
            reportRequest["endTime"],
            selectFields,
            reportRequest["condDict"],
            reportRequest["groupingFields"],
            {"convertToGranularity": "average"},
        )
        if not retVal["OK"]:
            return retVal

        dataDict, granularity = retVal["Value"]
        self.stripDataField(dataDict, 0)

        return S_OK({"data": dataDict, "granularity": granularity})

    def _plotPacketLossRate(self, reportRequest, plotInfo, filename):

        # prepare custom scale (10,20,...,100)
        scale_data = dict(zip(range(0, 101), range(100, -1, -1)))
        scale_ticks = list(range(0, 101, 10))

        metadata = {
            "title": "Packet loss rate by %s" % reportRequest["grouping"],
            "starttime": reportRequest["startTime"],
            "endtime": reportRequest["endTime"],
            "span": plotInfo["granularity"],
            "graph_size": "large",
            "reverse_labels": True,
            "scale_data": scale_data,
            "scale_ticks": scale_ticks,
        }
        return self._generateQualityPlot(filename, plotInfo["data"], metadata)

    _reportMagnifiedPacketLossRateName = "Packet loss rate (magnified)"

    def _reportMagnifiedPacketLossRate(self, reportRequest):

        selectFields = (
            self._getSelectStringForGrouping(reportRequest["groupingFields"]) + ", %s, %s, %s, %s",
            reportRequest["groupingFields"][1]
            + [
                "startTime",
                "bucketLength",
                "100 - IF(SUM(PacketLossRate)/SUM(entriesInBucket)*10 > 100, 100",
                "SUM(PacketLossRate)/SUM(entriesInBucket)*10), 100",
            ],
        )
        retVal = self._getTimedData(
            reportRequest["startTime"],
            reportRequest["endTime"],
            selectFields,
            reportRequest["condDict"],
            reportRequest["groupingFields"],
            {"convertToGranularity": "average"},
        )
        if not retVal["OK"]:
            return retVal

        dataDict, granularity = retVal["Value"]
        self.stripDataField(dataDict, 0)

        return S_OK({"data": dataDict, "granularity": granularity})

    def _plotMagnifiedPacketLossRate(self, reportRequest, plotInfo, filename):

        # prepare custom scale (1..10, 100)
        boundaries = list(np.arange(0, 10, 0.1))
        boundaries.extend(range(10, 110, 10))
        values = list(np.arange(100, 0, -1))
        values.extend([0] * 10)

        scale_data = dict(zip(boundaries, values))
        scale_ticks = list(range(0, 11))
        scale_ticks.append(100)

        metadata = {
            "title": "Magnified packet loss rate by %s" % reportRequest["grouping"],
            "starttime": reportRequest["startTime"],
            "endtime": reportRequest["endTime"],
            "span": plotInfo["granularity"],
            "reverse_labels": True,
            "graph_size": "large",
            "scale_data": scale_data,
            "scale_ticks": scale_ticks,
        }
        return self._generateQualityPlot(filename, plotInfo["data"], metadata)

    _reportAverageOneWayDelayName = "One-way delay (average)"

    def _reportAverageOneWayDelay(self, reportRequest):

        selectFields = (
            self._getSelectStringForGrouping(reportRequest["groupingFields"]) + ", %s, %s, SUM(%s)/SUM(%s)",
            reportRequest["groupingFields"][1] + ["startTime", "bucketLength", "OneWayDelay", "entriesInBucket"],
        )
        retVal = self._getTimedData(
            reportRequest["startTime"],
            reportRequest["endTime"],
            selectFields,
            reportRequest["condDict"],
            reportRequest["groupingFields"],
            {"convertToGranularity": "average"},
        )
        if not retVal["OK"]:
            return retVal

        dataDict, granularity = retVal["Value"]
        self.stripDataField(dataDict, 0)
        dataDict = self._fillWithZero(granularity, reportRequest["startTime"], reportRequest["endTime"], dataDict)

        return S_OK({"data": dataDict, "granularity": granularity, "unit": "ms"})

    def _plotAverageOneWayDelay(self, reportRequest, plotInfo, filename):
        metadata = {
            "title": "One-way delay by %s" % reportRequest["grouping"],
            "ylabel": plotInfo["unit"],
            "starttime": reportRequest["startTime"],
            "endtime": reportRequest["endTime"],
            "graph_size": "large",
            "span": plotInfo["granularity"],
            "sort_labels": "avg_nozeros",
            "legend_unit": plotInfo["unit"],
        }
        return self._generateStackedLinePlot(filename, plotInfo["data"], metadata)

    _reportJitterName = "Jitter"

    def _reportJitter(self, reportRequest):

        selectFields = (
            self._getSelectStringForGrouping(reportRequest["groupingFields"]) + ", %s, %s, SUM(%s)/SUM(%s)",
            reportRequest["groupingFields"][1] + ["startTime", "bucketLength", "Jitter", "entriesInBucket"],
        )
        retVal = self._getTimedData(
            reportRequest["startTime"],
            reportRequest["endTime"],
            selectFields,
            reportRequest["condDict"],
            reportRequest["groupingFields"],
            {"convertToGranularity": "average"},
        )
        if not retVal["OK"]:
            return retVal

        dataDict, granularity = retVal["Value"]
        self.stripDataField(dataDict, 0)
        dataDict = self._fillWithZero(granularity, reportRequest["startTime"], reportRequest["endTime"], dataDict)

        return S_OK({"data": dataDict, "granularity": granularity, "unit": "ms"})

    def _plotJitter(self, reportRequest, plotInfo, filename):
        metadata = {
            "title": "Jitter by %s" % reportRequest["grouping"],
            "ylabel": plotInfo["unit"],
            "starttime": reportRequest["startTime"],
            "endtime": reportRequest["endTime"],
            "graph_size": "large",
            "span": plotInfo["granularity"],
            "sort_labels": "avg_nozeros",
            "legend_unit": plotInfo["unit"],
        }
        return self._generateStackedLinePlot(filename, plotInfo["data"], metadata)

    _reportJitterDelayRatioName = "Jitter/Delay"

    def _reportJitterDelayRatio(self, reportRequest):

        selectFields = (
            self._getSelectStringForGrouping(reportRequest["groupingFields"]) + ", %s, %s, SUM(%s)/SUM(%s)",
            reportRequest["groupingFields"][1] + ["startTime", "bucketLength", "Jitter", "OneWayDelay"],
        )
        retVal = self._getTimedData(
            reportRequest["startTime"],
            reportRequest["endTime"],
            selectFields,
            reportRequest["condDict"],
            reportRequest["groupingFields"],
            {"convertToGranularity": "average"},
        )
        if not retVal["OK"]:
            return retVal

        dataDict, granularity = retVal["Value"]
        self.stripDataField(dataDict, 0)
        dataDict = self._fillWithZero(granularity, reportRequest["startTime"], reportRequest["endTime"], dataDict)

        return S_OK({"data": dataDict, "granularity": granularity})

    def _plotJitterDelayRatio(self, reportRequest, plotInfo, filename):
        metadata = {
            "title": "Jitter over one-way delay by %s" % reportRequest["grouping"],
            "ylabel": "",
            "starttime": reportRequest["startTime"],
            "endtime": reportRequest["endTime"],
            "graph_size": "large",
            "span": plotInfo["granularity"],
            "sort_labels": "avg_nozeros",
            "legend_unit": "",
        }
        return self._generateStackedLinePlot(filename, plotInfo["data"], metadata)



""" Module that expose the base class for DIRAC Clients.

    This class exposes possible RPC calls, given a url of a service.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import ast
from functools import partial

try:
    from functools import partialmethod
except ImportError:

    class partialmethod(partial):
        def __get__(self, instance, owner):
            if instance is None:
                return self
            return partial(self.func, instance, *(self.args or ()), **(self.keywords or {}))


import importlib_resources
import six

from DIRAC.Core.Tornado.Client.ClientSelector import RPCClientSelector
from DIRAC.Core.Tornado.Client.TornadoClient import TornadoClient
from DIRAC.Core.Utilities.Extensions import extensionsByPriority
from DIRAC.Core.Utilities.Decorators import deprecated
from DIRAC.Core.DISET import DEFAULT_RPC_TIMEOUT


class partialmethodWithDoc(partialmethod):
    """Extension of meth:`functools.partialmethod` that preserves docstrings"""

    def __get__(self, instance, owner):
        func = super(partialmethodWithDoc, self).__get__(instance, owner)
        func.__doc__ = self.__doc__
        return func


class Client(object):
    """Simple class to redirect unknown actions directly to the server. Arguments
    to the constructor are passed to the RPCClient constructor as they are.
    Some of them can however be overwritten at each call (url and timeout).

    - The self.serverURL member should be set by the inheriting class
    """

    # Default https (RPC)Client
    httpsClient = TornadoClient

    def __init__(self, **kwargs):
        """C'tor.

        :param kwargs: just stored as an attribute and passed when creating
                      the RPCClient
        """
        self.serverURL = kwargs.pop("url", None)
        self.__kwargs = kwargs
        self.timeout = DEFAULT_RPC_TIMEOUT

    def __getattr__(self, name):
        """Store the attribute asked and call executeRPC.
        This means that Client should not be shared between threads !
        """
        # This allows the dir() method to work as well as tab completion in ipython
        if name == "__dir__":
            return super(Client, self).__getattr__()  # pylint: disable=no-member
        return partial(self.executeRPC, call=name)

    def setServer(self, url):
        """Set the server URL used by default

        :param url: url of the service
        """
        self.serverURL = url

    def getClientKWArgs(self):
        """Returns a copy of the connection arguments"""
        return dict(self.__kwargs)

    def getServer(self):
        """Getter for the server url. Useful ?"""
        return self.serverURL

    @property
    @deprecated("To be removed once we're sure self.call has been removed")
    def call(self):
        raise NotImplementedError("This should be unreachable")

    def executeRPC(self, *parms, **kws):
        """This method extracts some parameters from kwargs that
        are used as parameter of the constructor or RPCClient.
        Unfortunately, only a few of all the available
        parameters of BaseClient are exposed.

        :param rpc: if an RPC client is passed, use that one
        :param timeout: we can change the timeout on a per call bases. Default is self.timeout
        :param url: We can specify which url to use
        """
        toExecute = kws.pop("call")
        # Check whether 'rpc' keyword is specified
        rpc = kws.pop("rpc", False)
        # Check whether the 'timeout' keyword is specified
        timeout = kws.pop("timeout", self.timeout)
        # Check whether the 'url' keyword is specified
        url = kws.pop("url", "")
        # Create the RPCClient
        rpcClient = self._getRPC(rpc=rpc, url=url, timeout=timeout)
        # Execute the method
        return getattr(rpcClient, toExecute)(*parms)

    def _getRPC(self, rpc=None, url="", timeout=None):
        """Return an RPCClient object constructed following the attributes.

        :param rpc: if set, returns this object
        :param url: url of the service. If not set, use self.serverURL
        :param timeout: timeout of the call. If not given, self.timeout will be used
        """
        if not rpc:
            if not url:
                url = self.serverURL

            if not timeout:
                timeout = self.timeout

            self.__kwargs["timeout"] = timeout
            rpc = RPCClientSelector(url, httpsClient=self.httpsClient, **self.__kwargs)
        return rpc


def createClient(serviceName):
    """Decorator to expose the service functions automatically in the Client.

    :param str serviceName: system/service. e.g. WorkloadManagement/JobMonitoring
    """
    systemName, handlerName = serviceName.split("/")
    handlerModuleName = handlerName + "Handler"
    # by convention they are the same
    handlerClassName = handlerModuleName
    handlerClassPath = "%sSystem.Service.%s.%s" % (systemName, handlerModuleName, handlerClassName)

    def genFunc(funcName, arguments, handlerClassPath, doc):
        """Create a function with *funcName* taking *arguments*."""
        doc = "" if doc is None else doc
        funcDocString = "%s(%s, **kwargs)\n" % (funcName, ", ".join(arguments))
        # do not describe self or cls in the parameter description
        if arguments and arguments[0] in ("self", "cls"):
            arguments = arguments[1:]

        # Create the actual functions, with or without arguments, **kwargs can be: rpc, timeout, url
        func = partialmethodWithDoc(Client.executeRPC, call=funcName)
        func.__doc__ = funcDocString + doc
        func.__doc__ += "\n\nAutomatically created for the service function "
        func.__doc__ += ":func:`~%s.export_%s`" % (handlerClassPath, funcName)
        # add description for parameters, if that is not already done for the docstring of function in the service
        if arguments and ":param " not in doc:
            func.__doc__ += "\n\n"
            func.__doc__ += "\n".join(":param %s: %s" % (par, par) for par in arguments)
        return func

    def addFunctions(clientCls):
        """Add the functions to the decorated class."""
        attrDict = dict(clientCls.__dict__)
        for extension in extensionsByPriority():
            try:
                path = importlib_resources.path(
                    "%s.%sSystem.Service" % (extension, systemName),
                    "%s.py" % handlerModuleName,
                )
                fullHandlerClassPath = "%s.%s" % (extension, handlerClassPath)
                with path as fp:
                    handlerAst = ast.parse(fp.read_text(), str(path))
            except (ImportError, OSError):
                continue

            # loop over all the nodes (classes, functions, imports) in the handlerModule
            for node in ast.iter_child_nodes(handlerAst):
                # find only a class with the name of the handlerClass
                if not (isinstance(node, ast.ClassDef) and node.name == handlerClassName):
                    continue
                for member in ast.iter_child_nodes(node):
                    # only look at functions
                    if not isinstance(member, ast.FunctionDef):
                        continue
                    if not member.name.startswith("export_"):
                        continue
                    funcName = member.name[len("export_") :]
                    if funcName in attrDict:
                        continue
                    if six.PY2:
                        arguments = [a.id for a in member.args.args]
                    else:
                        arguments = [a.arg for a in member.args.args]
                    # add the implementation of the function to the class attributes
                    attrDict[funcName] = genFunc(funcName, arguments, fullHandlerClassPath, ast.get_docstring(member))

        return type(clientCls.__name__, clientCls.__bases__, attrDict)

    return addFunctions


def executeRPCStub(rpcStub):
    """
    Playback a stub with the correct client (https or dips)
    """
    baseStub, methName, args = rpcStub
    url, callParams = baseStub
    # Make a copy to update it
    stub = dict(callParams)
    stub["url"] = url
    # Generate a RPCClient with the same parameters
    client = Client(**stub)
    # Get a functor to execute the RPC call
    rpcFunc = getattr(client, methName)
    # Reproduce the call
    return rpcFunc(*args)



""" CacheFeederAgent

  This agent feeds the Cache tables with the outputs of the cache commands.

.. literalinclude:: ../ConfigTemplate.cfg
  :start-after: ##BEGIN CacheFeederAgent
  :end-before: ##END
  :dedent: 2
  :caption: CacheFeederAgent options
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC import S_OK
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.Core.LCG.GOCDBClient import GOCDBClient
from DIRAC.Core.Utilities.ObjectLoader import ObjectLoader
from DIRAC.AccountingSystem.Client.ReportsClient import ReportsClient
from DIRAC.WorkloadManagementSystem.Client.WMSAdministratorClient import WMSAdministratorClient
from DIRAC.ResourceStatusSystem.Command import CommandCaller
from DIRAC.WorkloadManagementSystem.Client.PilotManagerClient import PilotManagerClient

AGENT_NAME = "ResourceStatus/CacheFeederAgent"


class CacheFeederAgent(AgentModule):
    """
    The CacheFeederAgent feeds the cache tables for the client and the accounting.
    It runs periodically a set of commands, and stores it's results on the
    tables.
    """

    def __init__(self, *args, **kwargs):

        AgentModule.__init__(self, *args, **kwargs)

        self.commands = {}
        self.clients = {}

        self.cCaller = None
        self.rmClient = None

    def initialize(self):
        """Define the commands to be executed, and instantiate the clients that will be used."""

        res = ObjectLoader().loadObject("DIRAC.ResourceStatusSystem.Client.ResourceStatusClient")
        if not res["OK"]:
            self.log.error("Failed to load ResourceStatusClient class: %s" % res["Message"])
            return res
        rsClass = res["Value"]

        res = ObjectLoader().loadObject("DIRAC.ResourceStatusSystem.Client.ResourceManagementClient")
        if not res["OK"]:
            self.log.error("Failed to load ResourceManagementClient class: %s" % res["Message"])
            return res
        rmClass = res["Value"]

        self.commands["Downtime"] = [{"Downtime": {}}]
        self.commands["GOCDBSync"] = [{"GOCDBSync": {}}]
        self.commands["FreeDiskSpace"] = [{"FreeDiskSpace": {}}]

        # PilotsCommand
        self.commands["Pilot"] = [
            {"Pilot": {"element": "Site", "siteName": None}},
            {"Pilot": {"element": "Resource", "siteName": None}},
        ]

        # FIXME: do not forget about hourly vs Always ...etc
        # AccountingCacheCommand
        #    self.commands[ 'AccountingCache' ] = [
        #                                          {'SuccessfullJobsBySiteSplitted'    :{'hours' :24, 'plotType' :'Job' }},
        #                                          {'FailedJobsBySiteSplitted'         :{'hours' :24, 'plotType' :'Job' }},
        #                                          {'SuccessfullPilotsBySiteSplitted'  :{'hours' :24, 'plotType' :'Pilot' }},
        #                                          {'FailedPilotsBySiteSplitted'       :{'hours' :24, 'plotType' :'Pilot' }},
        #                                          {'SuccessfullPilotsByCESplitted'    :{'hours' :24, 'plotType' :'Pilot' }},
        #                                          {'FailedPilotsByCESplitted'         :{'hours' :24, 'plotType' :'Pilot' }},
        #                                          {'RunningJobsBySiteSplitted'        :{'hours' :24, 'plotType' :'Job' }},
        # #                                          {'RunningJobsBySiteSplitted'        :{'hours' :168, 'plotType' :'Job' }},
        # #                                          {'RunningJobsBySiteSplitted'        :{'hours' :720, 'plotType' :'Job' }},
        # #                                          {'RunningJobsBySiteSplitted'        :{'hours' :8760, 'plotType' :'Job' }},
        #                                          ]

        # VOBOXAvailability
        #    self.commands[ 'VOBOXAvailability' ] = [
        #                                            { 'VOBOXAvailability' : {} }
        #

        # Reuse clients for the commands
        self.clients["GOCDBClient"] = GOCDBClient()
        self.clients["ReportsClient"] = ReportsClient()
        self.clients["ResourceStatusClient"] = rsClass()
        self.clients["ResourceManagementClient"] = rmClass()
        self.clients["WMSAdministrator"] = WMSAdministratorClient()
        self.clients["Pilots"] = PilotManagerClient()

        self.cCaller = CommandCaller

        return S_OK()

    def loadCommand(self, commandModule, commandDict):
        """Loads and executes commands.

        :param str commandModule: Name of the command (e.g. 'Downtime')
        :param dict commandDict: dictionary of {'CommandClass':{arguments}}
        """

        commandName = list(commandDict)[0]
        commandArgs = commandDict[commandName]

        commandTuple = ("%sCommand" % commandModule, "%sCommand" % commandName)
        commandObject = self.cCaller.commandInvocation(commandTuple, pArgs=commandArgs, clients=self.clients)

        if not commandObject["OK"]:
            self.log.error("Error initializing %s" % commandName)
            return commandObject
        commandObject = commandObject["Value"]

        # Set master mode
        commandObject.masterMode = True

        self.log.info("%s/%s" % (commandModule, commandName))

        return S_OK(commandObject)

    def execute(self):
        """Just executes, via `loadCommand`, the commands in self.commands one after the other"""

        for commandModule, commandList in self.commands.items():

            self.log.info("%s module initialization" % commandModule)

            for commandDict in commandList:

                commandObject = self.loadCommand(commandModule, commandDict)
                if not commandObject["OK"]:
                    self.log.error(commandObject["Message"])
                    continue
                commandObject = commandObject["Value"]

                try:
                    results = commandObject.doCommand()
                    if not results["OK"]:
                        self.log.error("Failed to execute command", "%s: %s" % (commandModule, results["Message"]))
                        continue
                    results = results["Value"]
                    if not results:
                        self.log.info("Empty results")
                        continue
                    self.log.verbose("Command OK Results")
                    self.log.verbose(results)
                except Exception as excp:  # pylint: disable=broad-except
                    self.log.exception("Failed to execute command, with exception: %s" % commandModule, lException=excp)

        return S_OK()



""" ConfigurationData module is the base for cfg files management
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
import os.path
import zlib
import zipfile
import six
from six.moves import _thread as thread
import time
import DIRAC

from diraccfg import CFG
from DIRAC.Core.Utilities.File import mkDir
from DIRAC.Core.Utilities import List, Time
from DIRAC.Core.Utilities.ReturnValues import S_OK, S_ERROR
from DIRAC.Core.Utilities.LockRing import LockRing
from DIRAC.FrameworkSystem.Client.Logger import gLogger

__RCSID__ = "$Id$"


class ConfigurationData(object):
    def __init__(self, loadDefaultCFG=True):
        envVar = os.environ.get("DIRAC_FEWER_CFG_LOCKS", "no").lower()
        self.__locksEnabled = envVar not in ("y", "yes", "t", "true", "on", "1")
        if self.__locksEnabled:
            lr = LockRing()
            self.threadingEvent = lr.getEvent()
            self.threadingEvent.set()
            self.threadingLock = lr.getLock()
            self.runningThreadsNumber = 0

        self.__compressedConfigurationData = None
        self.configurationPath = "/DIRAC/Configuration"
        self.backupsDir = os.path.join(DIRAC.rootPath, "etc", "csbackup")
        self._isService = False
        self.localCFG = CFG()
        self.remoteCFG = CFG()
        self.mergedCFG = CFG()
        self.remoteServerList = []
        if loadDefaultCFG:
            defaultCFGFile = os.path.join(DIRAC.rootPath, "etc", "dirac.cfg")
            gLogger.debug("dirac.cfg should be at", "%s" % defaultCFGFile)
            retVal = self.loadFile(defaultCFGFile)
            if not retVal["OK"]:
                gLogger.warn("Can't load %s file" % defaultCFGFile)
        self.sync()

    def getBackupDir(self):
        return self.backupsDir

    def sync(self):
        gLogger.debug("Updating configuration internals")
        self.mergedCFG = self.remoteCFG.mergeWith(self.localCFG)
        self.remoteServerList = []
        localServers = self.extractOptionFromCFG(
            "%s/Servers" % self.configurationPath, self.localCFG, disableDangerZones=True
        )
        if localServers:
            self.remoteServerList.extend(List.fromChar(localServers, ","))
        remoteServers = self.extractOptionFromCFG(
            "%s/Servers" % self.configurationPath, self.remoteCFG, disableDangerZones=True
        )
        if remoteServers:
            self.remoteServerList.extend(List.fromChar(remoteServers, ","))
        self.remoteServerList = List.uniqueElements(self.remoteServerList)
        self.__compressedConfigurationData = None

    def loadFile(self, fileName):
        try:
            fileCFG = CFG()
            fileCFG.loadFromFile(fileName)
        except IOError:
            self.localCFG = self.localCFG.mergeWith(fileCFG)
            return S_ERROR("Can't load a cfg file '%s'" % fileName)
        return self.mergeWithLocal(fileCFG)

    def mergeWithLocal(self, extraCFG):
        self.lock()
        try:
            self.localCFG = self.localCFG.mergeWith(extraCFG)
            self.unlock()
            gLogger.debug("CFG merged")
        except Exception as e:
            self.unlock()
            return S_ERROR("Cannot merge with new cfg: %s" % str(e))
        self.sync()
        return S_OK()

    def loadRemoteCFGFromCompressedMem(self, data):
        if six.PY3 and isinstance(data, str):
            data = data.encode(errors="surrogateescape")
        sUncompressedData = zlib.decompress(data).decode()
        self.loadRemoteCFGFromMem(sUncompressedData)

    def loadRemoteCFGFromMem(self, data):
        self.lock()
        self.remoteCFG.loadFromBuffer(data)
        self.unlock()
        self.sync()

    def loadConfigurationData(self, fileName=False):
        name = self.getName()
        self.lock()
        try:
            if not fileName:
                fileName = "%s.cfg" % name
            if fileName[0] != "/":
                fileName = os.path.join(DIRAC.rootPath, "etc", fileName)
            self.remoteCFG.loadFromFile(fileName)
        except Exception as e:
            print(e)
        self.unlock()
        self.sync()

    def getCommentFromCFG(self, path, cfg=False):
        if not cfg:
            cfg = self.mergedCFG
        self.dangerZoneStart()
        try:
            levelList = [level.strip() for level in path.split("/") if level.strip() != ""]
            for section in levelList[:-1]:
                cfg = cfg[section]
            return self.dangerZoneEnd(cfg.getComment(levelList[-1]))
        except Exception:
            pass
        return self.dangerZoneEnd(None)

    def getSectionsFromCFG(self, path, cfg=False, ordered=False):
        if not cfg:
            cfg = self.mergedCFG
        self.dangerZoneStart()
        try:
            levelList = [level.strip() for level in path.split("/") if level.strip() != ""]
            for section in levelList:
                cfg = cfg[section]
            return self.dangerZoneEnd(cfg.listSections(ordered))
        except Exception:
            pass
        return self.dangerZoneEnd(None)

    def getOptionsFromCFG(self, path, cfg=False, ordered=False):
        if not cfg:
            cfg = self.mergedCFG
        self.dangerZoneStart()
        try:
            levelList = [level.strip() for level in path.split("/") if level.strip() != ""]
            for section in levelList:
                cfg = cfg[section]
            return self.dangerZoneEnd(cfg.listOptions(ordered))
        except Exception:
            pass
        return self.dangerZoneEnd(None)

    def extractOptionFromCFG(self, path, cfg=False, disableDangerZones=False):
        if not cfg:
            cfg = self.mergedCFG
        if not disableDangerZones:
            self.dangerZoneStart()
        try:
            levelList = [level.strip() for level in path.split("/") if level.strip() != ""]
            for section in levelList[:-1]:
                cfg = cfg[section]
            if levelList[-1] in cfg.listOptions():
                return self.dangerZoneEnd(cfg[levelList[-1]])
        except Exception:
            pass
        if not disableDangerZones:
            self.dangerZoneEnd()

    def setOptionInCFG(self, path, value, cfg=False, disableDangerZones=False):
        if not cfg:
            cfg = self.localCFG
        if not disableDangerZones:
            self.dangerZoneStart()
        try:
            levelList = [level.strip() for level in path.split("/") if level.strip() != ""]
            for section in levelList[:-1]:
                if section not in cfg.listSections():
                    cfg.createNewSection(section)
                cfg = cfg[section]
            cfg.setOption(levelList[-1], value)
        finally:
            if not disableDangerZones:
                self.dangerZoneEnd()
        self.sync()

    def deleteOptionInCFG(self, path, cfg=False):
        if not cfg:
            cfg = self.localCFG
        self.dangerZoneStart()
        try:
            levelList = [level.strip() for level in path.split("/") if level.strip() != ""]
            for section in levelList[:-1]:
                if section not in cfg.listSections():
                    return
                cfg = cfg[section]
            cfg.deleteKey(levelList[-1])
        finally:
            self.dangerZoneEnd()
        self.sync()

    def generateNewVersion(self):
        self.setVersion(Time.toString())
        self.sync()
        gLogger.info("Generated new version %s" % self.getVersion())

    def setVersion(self, version, cfg=False):
        if not cfg:
            cfg = self.remoteCFG
        self.setOptionInCFG("%s/Version" % self.configurationPath, version, cfg)

    def getVersion(self, cfg=False):
        if not cfg:
            cfg = self.remoteCFG
        value = self.extractOptionFromCFG("%s/Version" % self.configurationPath, cfg)
        if value:
            return value
        return "0"

    def getName(self):
        return self.extractOptionFromCFG("%s/Name" % self.configurationPath, self.mergedCFG)

    def exportName(self):
        return self.setOptionInCFG("%s/Name" % self.configurationPath, self.getName(), self.remoteCFG)

    def getRefreshTime(self):
        try:
            return int(self.extractOptionFromCFG("%s/RefreshTime" % self.configurationPath, self.mergedCFG))
        except Exception:
            return 300

    def getPropagationTime(self):
        try:
            return int(self.extractOptionFromCFG("%s/PropagationTime" % self.configurationPath, self.mergedCFG))
        except Exception:
            return 300

    def getSlavesGraceTime(self):
        try:
            return int(self.extractOptionFromCFG("%s/SlavesGraceTime" % self.configurationPath, self.mergedCFG))
        except Exception:
            return 600

    def mergingEnabled(self):
        try:
            val = self.extractOptionFromCFG("%s/EnableAutoMerge" % self.configurationPath, self.mergedCFG)
            return val.lower() in ("yes", "true", "y")
        except Exception:
            return False

    def getAutoPublish(self):
        value = self.extractOptionFromCFG("%s/AutoPublish" % self.configurationPath, self.localCFG)
        if value and value.lower() in ("no", "false", "n"):
            return False
        else:
            return True

    def getAutoSlaveSync(self):
        value = self.extractOptionFromCFG("%s/AutoSlaveSync" % self.configurationPath, self.localCFG)
        if value and value.lower() in ("no", "false", "n"):
            return False
        else:
            return True

    def getServers(self):
        return list(self.remoteServerList)

    def getConfigurationGateway(self):
        return self.extractOptionFromCFG("/DIRAC/Gateway", self.localCFG)

    def setServers(self, sServers):
        self.setOptionInCFG("%s/Servers" % self.configurationPath, sServers, self.remoteCFG)
        self.sync()

    def deleteLocalOption(self, optionPath):
        self.deleteOptionInCFG(optionPath, self.localCFG)

    def getMasterServer(self):
        return self.extractOptionFromCFG("%s/MasterServer" % self.configurationPath, self.remoteCFG)

    def setMasterServer(self, sURL):
        self.setOptionInCFG("%s/MasterServer" % self.configurationPath, sURL, self.remoteCFG)
        self.sync()

    def getCompressedData(self):
        if self.__compressedConfigurationData is None:
            self.__compressedConfigurationData = zlib.compress(str(self.remoteCFG).encode(), 9)
        return self.__compressedConfigurationData

    def isMaster(self):
        value = self.extractOptionFromCFG("%s/Master" % self.configurationPath, self.localCFG)
        if value and value.lower() in ("yes", "true", "y"):
            return True
        else:
            return False

    def getServicesPath(self):
        return "/Services"

    def setAsService(self):
        self._isService = True

    def isService(self):
        return self._isService

    def useServerCertificate(self):
        value = self.extractOptionFromCFG("/DIRAC/Security/UseServerCertificate")
        if value and value.lower() in ("y", "yes", "true"):
            return True
        return False

    def skipCACheck(self):
        value = self.extractOptionFromCFG("/DIRAC/Security/SkipCAChecks")
        if value and value.lower() in ("y", "yes", "true"):
            return True
        return False

    def dumpLocalCFGToFile(self, fileName):
        try:
            with open(fileName, "w") as fd:
                fd.write(str(self.localCFG))
            gLogger.verbose("Configuration file dumped", "'%s'" % fileName)
        except IOError:
            gLogger.error("Can't dump cfg file", "'%s'" % fileName)
            return S_ERROR("Can't dump cfg file '%s'" % fileName)
        return S_OK()

    def getRemoteCFG(self):
        return self.remoteCFG

    def getMergedCFGAsString(self):
        return str(self.mergedCFG)

    def dumpRemoteCFGToFile(self, fileName):
        with open(fileName, "w") as fd:
            fd.write(str(self.remoteCFG))

    def __backupCurrentConfiguration(self, backupName):
        configurationFilename = "%s.cfg" % self.getName()
        configurationFile = os.path.join(DIRAC.rootPath, "etc", configurationFilename)
        today = Time.date()
        backupPath = os.path.join(self.getBackupDir(), str(today.year), "%02d" % today.month)
        mkDir(backupPath)
        backupFile = os.path.join(backupPath, configurationFilename.replace(".cfg", ".%s.zip" % backupName))
        if os.path.isfile(configurationFile):
            gLogger.info("Making a backup of configuration in %s" % backupFile)
            try:
                with zipfile.ZipFile(backupFile, "w", zipfile.ZIP_DEFLATED) as zf:
                    zf.write(configurationFile, "%s.backup.%s" % (os.path.split(configurationFile)[1], backupName))
            except Exception:
                gLogger.exception()
                gLogger.error("Cannot backup configuration data file", "file %s" % backupFile)
        else:
            gLogger.warn("CS data file does not exist", configurationFile)

    def writeRemoteConfigurationToDisk(self, backupName=False):
        configurationFile = os.path.join(DIRAC.rootPath, "etc", "%s.cfg" % self.getName())
        try:
            with open(configurationFile, "w") as fd:
                fd.write(str(self.remoteCFG))
        except Exception as e:
            gLogger.fatal(
                "Cannot write new configuration to disk!", "file %s exception %s" % (configurationFile, repr(e))
            )
            return S_ERROR("Can't write cs file %s!: %s" % (configurationFile, repr(e).replace(",)", ")")))
        if backupName:
            self.__backupCurrentConfiguration(backupName)
        return S_OK()

    def setRemoteCFG(self, cfg, disableSync=False):
        self.remoteCFG = cfg.clone()
        if not disableSync:
            self.sync()

    def lock(self):
        """
        Locks Event to prevent further threads from reading.
        Stops current thread until no other thread is accessing.
        PRIVATE USE
        """
        if not self.__locksEnabled:
            return
        self.threadingEvent.clear()
        while self.runningThreadsNumber > 0:
            time.sleep(0.1)

    def unlock(self):
        """
        Unlocks Event.
        PRIVATE USE
        """
        if not self.__locksEnabled:
            return
        self.threadingEvent.set()

    def dangerZoneStart(self):
        """
        Start of danger zone. This danger zone may be or may not be a mutual exclusion zone.
        Counter is maintained to know how many threads are inside and be able to enable and disable mutual exclusion.
        PRIVATE USE
        """
        if not self.__locksEnabled:
            return
        self.threadingEvent.wait()
        self.threadingLock.acquire()
        self.runningThreadsNumber += 1
        try:
            self.threadingLock.release()
        except thread.error:
            pass

    def dangerZoneEnd(self, returnValue=None):
        """
        End of danger zone.
        PRIVATE USE
        """
        if not self.__locksEnabled:
            return returnValue
        self.threadingLock.acquire()
        self.runningThreadsNumber -= 1
        try:
            self.threadingLock.release()
        except thread.error:
            pass
        return returnValue



""" Test class for agents
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# pylint: disable=protected-access, missing-docstring, invalid-name, line-too-long

# imports
import datetime

import pytest
from mock import MagicMock

from DIRAC.TransformationSystem.Client import TransformationFilesStatus

# sut
from DIRAC.TransformationSystem.Agent.TaskManagerAgentBase import TaskManagerAgentBase
from DIRAC.TransformationSystem.Agent.TransformationAgent import TransformationAgent

mockAM = MagicMock()


tc_mock = MagicMock()
tm_mock = MagicMock()
clients = {"TransformationClient": tc_mock, "TaskManager": tm_mock}

transDict = {"TransformationID": 1, "Operations": ["op1", "op2"], "Body": "veryBigBody"}
tasks = {
    "OK": True,
    "Value": [
        {
            "CreationTime": None,
            "ExternalID": "1",
            "ExternalStatus": "Reserved",
            "LastUpdateTime": None,
            "RunNumber": 0,
            "TargetSE": "Unknown",
            "TaskID": 1,
            "TransformationID": 101,
        },
        {
            "CreationTime": datetime.datetime(2014, 7, 21, 14, 19, 3),
            "ExternalID": "0",
            "ExternalStatus": "Reserved",
            "LastUpdateTime": datetime.datetime(2014, 7, 21, 14, 19, 3),
            "RunNumber": 0,
            "TargetSE": "Unknown",
            "TaskID": 2,
            "TransformationID": 101,
        },
    ],
}
sOk = {"OK": True, "Value": []}
sError = {"OK": False, "Message": "a mess"}


@pytest.mark.parametrize(
    "tcMockReturnValue, tmMockGetSubmittedTaskStatusReturnvalue, expected",
    [
        (sError, {"OK": True}, False),  # errors
        (sOk, {"OK": True}, True),  # no tasks
        (tasks, sError, False),  # tasks, fail in update
        (tasks, {"OK": True, "Value": {}}, True),  # tasks, nothing to update
        (tasks, {"OK": True, "Value": {"Running": [1, 2], "Done": [3]}}, True),  # tasks, to update, no errors
    ],
)
def test_updateTaskStatusSuccess(mocker, tcMockReturnValue, tmMockGetSubmittedTaskStatusReturnvalue, expected):
    mocker.patch("DIRAC.TransformationSystem.Agent.TaskManagerAgentBase.AgentModule", side_effect=mockAM)
    mocker.patch("DIRAC.TransformationSystem.Agent.TaskManagerAgentBase.FileReport", side_effect=MagicMock())
    mocker.patch(
        "DIRAC.TransformationSystem.Agent.TaskManagerAgentBase.TaskManagerAgentBase.am_getOption", side_effect=mockAM
    )
    tmab = TaskManagerAgentBase()
    tc_mock.getTransformationTasks.return_value = tcMockReturnValue
    tm_mock.getSubmittedTaskStatus.return_value = tmMockGetSubmittedTaskStatusReturnvalue
    res = tmab.updateTaskStatus(transDict, clients)
    assert res["OK"] == expected


@pytest.mark.parametrize(
    "tcMockGetTransformationFilesReturnValue, tmMockGetSubmittedFileStatusReturnValue, expected",
    [
        (sError, None, False),  # errors
        (sOk, None, True),  # no files
        ({"OK": True, "Value": [{"file1": "boh", "TaskID": 1}]}, sError, False),  # files, failing to update
        ({"OK": True, "Value": [{"file1": "boh", "TaskID": 1}]}, sOk, True),  # files, nothing to update
        (
            {"OK": True, "Value": [{"file1": "boh", "TaskID": 1}]},
            {"OK": True, "Value": {"file1": "OK", "file2": "NOK"}},
            True,
        ),  # files, something to update
    ],
)
def test_updateFileStatusSuccess(
    mocker, tcMockGetTransformationFilesReturnValue, tmMockGetSubmittedFileStatusReturnValue, expected
):
    mocker.patch("DIRAC.TransformationSystem.Agent.TaskManagerAgentBase.AgentModule", side_effect=mockAM)
    mocker.patch("DIRAC.TransformationSystem.Agent.TaskManagerAgentBase.FileReport", side_effect=MagicMock())
    tmab = TaskManagerAgentBase()
    tc_mock.getTransformationFiles.return_value = tcMockGetTransformationFilesReturnValue
    tm_mock.getSubmittedFileStatus.return_value = tmMockGetSubmittedFileStatusReturnValue
    res = tmab.updateFileStatus(transDict, clients)
    assert res["OK"] == expected


@pytest.mark.parametrize(
    ", ".join(
        [
            "tcMockGetTransformationTasksReturnValue",
            "tmMockUpdateTransformationReservedTasksReturnValue",
            "tcMockSetTaskStatusAndWmsIDReturnValue",
            "expected",
        ]
    ),
    [
        (sError, None, None, False),  # errors getting
        (sOk, None, None, True),  # no tasks
        (tasks, sError, None, False),  # tasks, failing to update
        (
            tasks,
            {"OK": True, "Value": {"NoTasks": [], "TaskNameIDs": {"1_1": 123, "2_1": 456}}},
            sError,
            False,
        ),  # tasks, something to update, fail
        (
            tasks,
            {"OK": True, "Value": {"NoTasks": ["3_4", "5_6"], "TaskNameIDs": {"1_1": 123, "2_1": 456}}},
            {"OK": True},
            True,
        ),
    ],
)  # tasks, something to update, no fail
def test_checkReservedTasks(
    mocker,
    tcMockGetTransformationTasksReturnValue,
    tmMockUpdateTransformationReservedTasksReturnValue,
    tcMockSetTaskStatusAndWmsIDReturnValue,
    expected,
):
    mocker.patch("DIRAC.TransformationSystem.Agent.TaskManagerAgentBase.AgentModule", side_effect=mockAM)
    mocker.patch("DIRAC.TransformationSystem.Agent.TaskManagerAgentBase.FileReport", side_effect=MagicMock())
    tmab = TaskManagerAgentBase()
    tc_mock.getTransformationTasks.return_value = tcMockGetTransformationTasksReturnValue
    tm_mock.updateTransformationReservedTasks.return_value = tmMockUpdateTransformationReservedTasksReturnValue
    tc_mock.setTaskStatusAndWmsID.return_value = tcMockSetTaskStatusAndWmsIDReturnValue
    res = tmab.checkReservedTasks(transDict, clients)
    assert res["OK"] == expected


transDict = {
    "TransformationID": 1,
    "Operations": ["op1", "op2"],
    "Body": "veryBigBody",
    "Owner": "prodMan",
    "OwnerDN": "/ca=man/user=prodMan",
    "OwnerGroup": "prodMans",
}
sOkJobDict = {"OK": True, "Value": {"JobDictionary": {123: "foo", 456: "bar"}}}
sOkJobs = {"OK": True, "Value": {123: "foo", 456: "bar"}}


@pytest.mark.parametrize(
    ", ".join(
        [
            "tcMockGetTasksToSubmitReturnValue",
            "tmMockPrepareTransformationTasksReturnValue",
            "tmMockSubmitTransformationTasksReturnValue",
            "tmMockUpdateDBAfterTaskSubmissionReturnValue",
            "expected",
        ]
    ),
    [
        (sError, None, None, None, False),  # errors getting
        ({"OK": True, "Value": {"JobDictionary": {}}}, None, None, None, True),  # no tasks
        (sOkJobDict, sError, None, None, False),  # tasks, errors
        (sOkJobDict, sOkJobs, sError, None, False),  # tasks, still errors
        (sOkJobDict, sOkJobs, sOk, sError, False),  # tasks, still errors
        (sOkJobDict, sOkJobs, sOk, sOk, True),
    ],
)  # tasks, no errors
def test_submitTasks(
    mocker,
    tcMockGetTasksToSubmitReturnValue,
    tmMockPrepareTransformationTasksReturnValue,
    tmMockSubmitTransformationTasksReturnValue,
    tmMockUpdateDBAfterTaskSubmissionReturnValue,
    expected,
):
    mocker.patch("DIRAC.TransformationSystem.Agent.TaskManagerAgentBase.AgentModule", side_effect=mockAM)
    mocker.patch("DIRAC.TransformationSystem.Agent.TaskManagerAgentBase.FileReport", side_effect=MagicMock())
    tmab = TaskManagerAgentBase()
    tc_mock.getTasksToSubmit.return_value = tcMockGetTasksToSubmitReturnValue
    tm_mock.prepareTransformationTasks.return_value = tmMockPrepareTransformationTasksReturnValue
    tm_mock.submitTransformationTasks.return_value = tmMockSubmitTransformationTasksReturnValue
    tm_mock.updateDBAfterTaskSubmission.return_value = tmMockUpdateDBAfterTaskSubmissionReturnValue
    res = tmab.submitTasks(transDict, clients)
    assert res["OK"] == expected
    tmab.maxParametricJobs = 10
    tmab.bulkSubmissionFlag = True
    res = tmab.submitTasks(transDict, clients)
    assert res["OK"] == expected


# TransformationAgent

goodFiles = {
    "OK": True,
    "Value": [
        {
            "ErrorCount": 1,
            "FileID": 17990660,
            "InsertedTime": datetime.datetime(2012, 3, 15, 17, 5, 50),
            "LFN": "/00012574_00000239_1.charmcompleteevent.dst",
            "LastUpdate": datetime.datetime(2012, 3, 16, 23, 43, 26),
            "RunNumber": 90269,
            "Status": TransformationFilesStatus.UNUSED,
            "TargetSE": "Unknown",
            "TaskID": "222",
            "TransformationID": 17042,
            "UsedSE": "CERN-DST,IN2P3-DST,PIC-DST,RAL-DST",
        },
        {
            "ErrorCount": 1,
            "FileID": 17022945,
            "InsertedTime": datetime.datetime(2012, 3, 15, 17, 5, 50),
            "LFN": "/00012574_00000119_1.charmcompleteevent.dst",
            "LastUpdate": datetime.datetime(2012, 3, 16, 23, 54, 59),
            "RunNumber": 90322,
            "Status": TransformationFilesStatus.UNUSED,
            "TargetSE": "Unknown",
            "TaskID": "82",
            "TransformationID": 17042,
            "UsedSE": "CERN-DST,CNAF-DST,RAL-DST,SARA-DST",
        },
    ],
}
noFiles = {"OK": True, "Value": []}


@pytest.mark.parametrize(
    "transDict, getTFiles, expected",
    [
        ({"TransformationID": 123, "Status": "Stopped", "Type": "Replication"}, goodFiles, True),
        ({"TransformationID": 123, "Status": "Stopped", "Type": "Removal"}, noFiles, True),
    ],
)
def test__getTransformationFiles(mocker, transDict, getTFiles, expected):
    mocker.patch("DIRAC.TransformationSystem.Agent.TransformationAgent.AgentModule", side_effect=mockAM)
    tc_mock.getTransformationFiles.return_value = getTFiles
    res = TransformationAgent()._getTransformationFiles(transDict, {"TransformationClient": tc_mock})
    assert res["OK"] == expected



""" Contains unit tests of NetworkAgent module
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import unittest
import sys

import DIRAC.AccountingSystem.Agent.NetworkAgent as module

from mock.mock import MagicMock

__RCSID__ = "$Id$"

MQURI1 = "mq.dirac.net::Topics::perfsonar.summary.packet-loss-rate"
MQURI2 = "mq.dirac.net::Queues::perfsonar.summary.histogram-owdelay"

ROOT_PATH = "/Resources/Sites"
SITE1 = "LCG.Dirac.net"
SITE2 = "LCG.DiracToRemove.net"
SITE3 = "VAC.DiracToAdd.org"
SITE1_HOST1 = "perfsonar.diracold.net"
SITE1_HOST2 = "perfsonar-to-disable.diracold.net"
SITE2_HOST1 = "perfsonar.diractoremove.net"
SITE3_HOST1 = "perfsonar.diractoadd.org"

INITIAL_CONFIG = {
    "%s/LCG/%s/Network/%s/Enabled" % (ROOT_PATH, SITE1, SITE1_HOST1): "True",
    "%s/LCG/%s/Network/%s/Enabled" % (ROOT_PATH, SITE1, SITE1_HOST2): "True",
    "%s/LCG/%s/Network/%s/Enabled" % (ROOT_PATH, SITE2, SITE2_HOST1): "True",
}

UPDATED_CONFIG = {
    "%s/LCG/%s/Network/%s/Enabled" % (ROOT_PATH, SITE1, SITE1_HOST1): "True",
    "%s/LCG/%s/Network/%s/Enabled" % (ROOT_PATH, SITE1, SITE1_HOST2): "False",
    "%s/LCG/%s/Network/%s/Enabled" % (ROOT_PATH, SITE3, SITE3_HOST1): "True",
}


class NetworkAgentSuccessTestCase(unittest.TestCase):
    """Test class to check success scenarios."""

    def setUp(self):

        # external dependencies
        module.datetime = MagicMock()

        # internal dependencies
        module.S_ERROR = MagicMock()
        module.S_OK = MagicMock()
        module.gLogger = MagicMock()
        module.AgentModule = MagicMock()
        module.Network = MagicMock()
        module.gConfig = MagicMock()
        module.CSAPI = MagicMock()
        module.createConsumer = MagicMock()

        # prepare test object
        module.NetworkAgent.__init__ = MagicMock(return_value=None)
        module.NetworkAgent.am_getOption = MagicMock(return_value=100)  # buffer timeout

        self.agent = module.NetworkAgent()
        self.agent.initialize()

    @classmethod
    def tearDownClass(cls):
        sys.modules.pop("DIRAC.AccountingSystem.Agent.NetworkAgent")

    def test_updateNameDictionary(self):

        module.gConfig.getConfigurationTree.side_effect = [
            {"OK": True, "Value": INITIAL_CONFIG},
            {"OK": True, "Value": UPDATED_CONFIG},
        ]

        # check if name dictionary is empty
        self.assertFalse(self.agent.nameDictionary)

        self.agent.updateNameDictionary()
        self.assertEqual(self.agent.nameDictionary[SITE1_HOST1], SITE1)
        self.assertEqual(self.agent.nameDictionary[SITE1_HOST2], SITE1)
        self.assertEqual(self.agent.nameDictionary[SITE2_HOST1], SITE2)

        self.agent.updateNameDictionary()
        self.assertEqual(self.agent.nameDictionary[SITE1_HOST1], SITE1)
        self.assertEqual(self.agent.nameDictionary[SITE3_HOST1], SITE3)

        # check if hosts were removed form dictionary
        self.assertRaises(KeyError, lambda: self.agent.nameDictionary[SITE1_HOST2])
        self.assertRaises(KeyError, lambda: self.agent.nameDictionary[SITE2_HOST1])

    def test_agentExecute(self):

        module.NetworkAgent.am_getOption.return_value = "%s, %s" % (MQURI1, MQURI2)
        module.gConfig.getConfigurationTree.return_value = {"OK": True, "Value": INITIAL_CONFIG}

        # first run
        result = self.agent.execute()
        self.assertTrue(result["OK"])

        # second run (simulate new messages)
        self.agent.messagesCount += 10
        result = self.agent.execute()
        self.assertTrue(result["OK"])

        # third run (no new messages - restart consumers)
        result = self.agent.execute()
        self.assertTrue(result["OK"])


if __name__ == "__main__":
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(NetworkAgentSuccessTestCase)
    testResult = unittest.TextTestRunner(verbosity=2).run(suite)



""" The TaskQueuesAgent will update TQs

.. literalinclude:: ../ConfigTemplate.cfg
  :start-after: ##BEGIN TaskQueuesAgent
  :end-before: ##END
  :dedent: 2
  :caption: TaskQueuesAgent options

"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC import S_OK
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.WorkloadManagementSystem.DB.TaskQueueDB import TaskQueueDB


class TaskQueuesAgent(AgentModule):
    """Agent for recalculating TQ shares"""

    def __init__(self, *args, **kwargs):
        """c'tor"""
        AgentModule.__init__(self, *args, **kwargs)

        # clients
        self.tqDB = None

    def initialize(self):
        """just initialize TQDB"""
        self.tqDB = TaskQueueDB()
        return S_OK()

    def execute(self):
        """calls TQDB.recalculateTQSharesForAll"""
        res = self.tqDB.recalculateTQSharesForAll()
        if not res["OK"]:
            self.log.error("Error recalculating TQ shares", res["Message"])

        return S_OK()



"""
  This module contains helper methods for accessing operational attributes or parameters of DMS objects

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import six
from DIRAC import gConfig, gLogger, S_OK, S_ERROR
from DIRAC.ConfigurationSystem.Client.Helpers.Path import cfgPath
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations

LOCAL = 1
PROTOCOL = LOCAL + 1
DOWNLOAD = PROTOCOL + 1


def resolveSEGroup(seGroupList, allSEs=None):
    """
    Resolves recursively a (list of) SEs that can be groupSEs

    :param seGroupList: list of SEs to resolve or comma-separated SEs
    :type seGroupList: list or string
    :param allSEs: if provided, list of all known SEs
    :type allSEs: list

    :return: list of resolved SEs or [] if error
    """
    if allSEs is None:
        res = gConfig.getSections("/Resources/StorageElements")
        if not res["OK"]:
            gLogger.fatal("Error getting list of SEs from CS", res["Message"])
            return []
        allSEs = res["Value"]
    seList = []
    if isinstance(seGroupList, six.string_types):
        seGroupList = [se.strip() for se in seGroupList.split(",") if se.strip()]
    for se in seGroupList:
        seConfig = gConfig.getValue("/Resources/StorageElementGroups/%s" % se, se)
        if seConfig != se:
            newSEs = [se1.strip() for se1 in seConfig.split(",") if se1.strip()]
            # print seList
        else:
            newSEs = [se]
        for se1 in list(newSEs):
            if se1 not in allSEs:
                # Here means se is not a group and is not an SE either, fatal!
                if se1 == se:
                    gLogger.fatal("%s is not a valid SE" % se1)
                    return []
                # If not an SE, it may be a group
                recursive = resolveSEGroup(se1, allSEs=allSEs)
                if not recursive:
                    return []
                newSEs.remove(se1)
                newSEs += recursive
        seList += newSEs

    return seList


def siteGridName(site):
    """Returns the Grid name for a site"""
    if not isinstance(site, six.string_types):
        return None
    siteSplit = site.split(".")
    if len(siteSplit) < 3:
        return None
    return siteSplit[0]


def siteCountryName(site):
    """Returns the Grid name for a site"""
    if not isinstance(site, six.string_types):
        return None
    siteSplit = site.split(".")
    if len(siteSplit) < 3:
        return None
    return site.split(".")[-1].lower()


def _getConnectionIndex(connectionLevel, default=None):
    """Converts a litteral connectivity to an integer value"""
    if connectionLevel is None:
        connectionLevel = default
    if isinstance(connectionLevel, six.integer_types):
        return connectionLevel
    if isinstance(connectionLevel, six.string_types):
        connectionLevel = connectionLevel.upper()
    return {"LOCAL": LOCAL, "PROTOCOL": PROTOCOL, "DOWNLOAD": DOWNLOAD}.get(connectionLevel)


class DMSHelpers(object):
    """
    This class is used to get information about sites, SEs and their interrelations
    """

    def __init__(self, vo=False):
        self.siteSEMapping = {}
        self.storageElementSet = set()
        self.siteSet = set()
        self.__opsHelper = Operations(vo=vo)
        self.failoverSEs = None
        self.archiveSEs = None
        self.notForJobSEs = None

    def getSiteSEMapping(self):
        """Returns a dictionary of all sites and their localSEs as a list, e.g.
        {'LCG.CERN.ch':['CERN-RAW','CERN-RDST',...]}
        """
        if self.siteSEMapping:
            return S_OK(self.siteSEMapping)

        # Get the list of SEs and keep a mapping of those using an Alias or a
        # BaseSE
        storageElements = gConfig.getSections("Resources/StorageElements")
        if not storageElements["OK"]:
            gLogger.warn("Problem retrieving storage elements", storageElements["Message"])
            return storageElements
        storageElements = storageElements["Value"]
        equivalentSEs = {}
        for se in storageElements:
            for option in ("BaseSE", "Alias"):
                originalSE = gConfig.getValue("Resources/StorageElements/%s/%s" % (se, option))
                if originalSE:
                    equivalentSEs.setdefault(originalSE, []).append(se)
                    break

        siteSEMapping = {}
        gridTypes = gConfig.getSections("Resources/Sites/")
        if not gridTypes["OK"]:
            gLogger.warn("Problem retrieving sections in /Resources/Sites", gridTypes["Message"])
            return gridTypes

        gridTypes = gridTypes["Value"]

        gLogger.debug("Grid Types are: %s" % (", ".join(gridTypes)))
        # Get a list of sites and their local SEs
        siteSet = set()
        storageElementSet = set()
        siteSEMapping[LOCAL] = {}
        for grid in gridTypes:
            result = gConfig.getSections("/Resources/Sites/%s" % grid)
            if not result["OK"]:
                gLogger.warn("Problem retrieving /Resources/Sites/%s section" % grid)
                return result
            sites = result["Value"]
            siteSet.update(sites)
            for site in sites:
                candidateSEs = gConfig.getValue("/Resources/Sites/%s/%s/SE" % (grid, site), [])
                if candidateSEs:
                    candidateSEs += [eqSE for se in candidateSEs for eqSE in equivalentSEs.get(se, [])]
                    siteSEMapping[LOCAL].setdefault(site, set()).update(candidateSEs)
                    storageElementSet.update(candidateSEs)

        # Add Sites from the SiteSEMappingByProtocol in the CS
        siteSEMapping[PROTOCOL] = {}
        cfgLocalSEPath = cfgPath("SiteSEMappingByProtocol")
        result = self.__opsHelper.getOptionsDict(cfgLocalSEPath)
        if result["OK"]:
            sites = result["Value"]
            for site in sites:
                candidates = set(self.__opsHelper.getValue(cfgPath(cfgLocalSEPath, site), []))
                ses = set(resolveSEGroup(candidates - siteSet)) | (candidates & siteSet)
                # If a candidate is a site, then all local SEs are eligible
                for candidate in ses & siteSet:
                    ses.remove(candidate)
                    ses.update(siteSEMapping[LOCAL][candidate])
                siteSEMapping[PROTOCOL].setdefault(site, set()).update(ses)

        # Add Sites from the SiteSEMappingByDownload in the CS, else
        # SiteLocalSEMapping (old convention)
        siteSEMapping[DOWNLOAD] = {}
        cfgLocalSEPath = cfgPath("SiteSEMappingByDownload")
        result = self.__opsHelper.getOptionsDict(cfgLocalSEPath)
        if not result["OK"]:
            cfgLocalSEPath = cfgPath("SiteLocalSEMapping")
            result = self.__opsHelper.getOptionsDict(cfgLocalSEPath)
        if result["OK"]:
            sites = result["Value"]
            for site in sites:
                candidates = set(self.__opsHelper.getValue(cfgPath(cfgLocalSEPath, site), []))
                ses = set(resolveSEGroup(candidates - siteSet)) | (candidates & siteSet)
                # If a candidate is a site, then all local SEs are eligible
                for candidate in ses & siteSet:
                    ses.remove(candidate)
                    ses.update(siteSEMapping[LOCAL][candidate])
                siteSEMapping[DOWNLOAD].setdefault(site, set()).update(ses)

        self.siteSEMapping = siteSEMapping
        # Add storage elements that may not be associated with a site
        result = gConfig.getSections("/Resources/StorageElements")
        if not result["OK"]:
            gLogger.warn("Problem retrieving /Resources/StorageElements section", result["Message"])
            return result
        self.storageElementSet = storageElementSet | set(result["Value"])
        self.siteSet = siteSet
        return S_OK(siteSEMapping)

    def getSites(self):
        """Get the list of known sites"""
        self.getSiteSEMapping()
        return sorted(self.siteSet)

    def getTiers(self, withStorage=False, tier=None):
        """Get the list of sites for a given (list of) Tier level"""
        sites = sorted(self.getShortSiteNames(withStorage=withStorage, tier=tier).values())
        if sites and isinstance(sites[0], list):
            # List of lists, flatten it
            sites = [s for sl in sites for s in sl]
        return sites

    def getShortSiteNames(self, withStorage=True, tier=None):
        """Create a directory of short site names pointing to full site names"""
        siteDict = {}
        result = self.getSiteSEMapping()
        if result["OK"]:
            for site in self.siteSEMapping[LOCAL] if withStorage else self.siteSet:
                grid, shortSite, _country = site.split(".")
                if isinstance(tier, six.integer_types) and (
                    grid != "LCG" or gConfig.getValue("/Resources/Sites/%s/%s/MoUTierLevel" % (grid, site), 999) != tier
                ):
                    continue
                if isinstance(tier, (list, tuple, dict, set)) and (
                    grid != "LCG"
                    or gConfig.getValue("/Resources/Sites/%s/%s/MoUTierLevel" % (grid, site), 999) not in tier
                ):
                    continue
                if withStorage or tier is not None:
                    siteDict[shortSite] = site
                else:
                    siteDict.setdefault(shortSite, []).append(site)
        return siteDict

    def getStorageElements(self):
        """Get the list of known SEs"""
        self.getSiteSEMapping()
        return sorted(self.storageElementSet)

    def isSEFailover(self, storageElement):
        """Is this SE a failover SE"""
        if self.failoverSEs is None:
            seList = resolveSEGroup(self.__opsHelper.getValue("DataManagement/SEsUsedForFailover", []))
            self.failoverSEs = resolveSEGroup(seList)
        # FIXME: remove string test at some point
        return storageElement in self.failoverSEs or (
            not self.failoverSEs
            and isinstance(storageElement, six.string_types)
            and "FAILOVER" in storageElement.upper()
        )

    def isSEForJobs(self, storageElement, checkSE=True):
        """Is this SE suitable for making jobs"""
        if checkSE:
            self.getSiteSEMapping()
            if storageElement not in self.storageElementSet:
                return False
        if self.notForJobSEs is None:
            seList = resolveSEGroup(self.__opsHelper.getValue("DataManagement/SEsNotToBeUsedForJobs", []))
            self.notForJobSEs = resolveSEGroup(seList)
        return storageElement not in self.notForJobSEs

    def isSEArchive(self, storageElement):
        """Is this SE an archive SE"""
        if self.archiveSEs is None:
            seList = resolveSEGroup(self.__opsHelper.getValue("DataManagement/SEsUsedForArchive", []))
            self.archiveSEs = resolveSEGroup(seList)
        # FIXME: remove string test at some point
        return storageElement in self.archiveSEs or (
            not self.archiveSEs and isinstance(storageElement, six.string_types) and "ARCHIVE" in storageElement.upper()
        )

    def getSitesForSE(self, storageElement, connectionLevel=None):
        """Get the (list of) sites for a given SE and a given connctivity"""
        connectionIndex = _getConnectionIndex(connectionLevel, default=DOWNLOAD)
        if connectionIndex == LOCAL:
            return self._getLocalSitesForSE(storageElement)
        if connectionIndex == PROTOCOL:
            return self.getProtocolSitesForSE(storageElement)
        if connectionIndex == DOWNLOAD:
            return self.getDownloadSitesForSE(storageElement)
        return S_ERROR("Unknown connection level")

    def getLocalSiteForSE(self, se):
        """Get the site at which the SE is"""
        sites = self._getLocalSitesForSE(se)
        if not sites["OK"]:
            return sites
        if not sites["Value"]:
            return S_OK(None)
        return S_OK(sites["Value"][0])

    def _getLocalSitesForSE(self, se):
        """Extract the list of sites that declare this SE"""
        mapping = self.getSiteSEMapping()
        if not mapping["OK"]:
            return mapping
        if se not in self.storageElementSet:
            return S_ERROR("Non-existing SE")
        mapping = mapping["Value"][LOCAL]
        sites = [site for site in mapping if se in mapping[site]]
        if len(sites) > 1 and self.__opsHelper.getValue("DataManagement/ForceSingleSitePerSE", True):
            return S_ERROR("SE is at more than one site")
        return S_OK(sites)

    def getProtocolSitesForSE(self, se):
        """Get sites that can access the SE by protocol"""
        mapping = self.getSiteSEMapping()
        if not mapping["OK"]:
            return mapping
        if se not in self.storageElementSet:
            return S_ERROR("Non-existing SE")
        mapping = mapping["Value"][PROTOCOL]
        sites = self._getLocalSitesForSE(se)
        if not sites["OK"]:
            return sites
        sites = set(sites["Value"])
        sites.update([site for site in mapping if se in mapping[site]])
        return S_OK(sorted(sites))

    def getDownloadSitesForSE(self, se):
        """Get the list of sites that are allowed to download files"""
        mapping = self.getSiteSEMapping()
        if not mapping["OK"]:
            return mapping
        if se not in self.storageElementSet:
            return S_ERROR("Non-existing SE")
        mapping = mapping["Value"][DOWNLOAD]
        sites = self.getProtocolSitesForSE(se)
        if not sites["OK"]:
            return sites
        sites = set(sites["Value"])
        sites.update([site for site in mapping if se in mapping[site]])
        return S_OK(sorted(sites))

    def getSEsForSite(self, site, connectionLevel=None):
        """Get all SEs accessible from a site, given a connectivity"""
        connectionIndex = _getConnectionIndex(connectionLevel, default=DOWNLOAD)
        if connectionIndex is None:
            return S_ERROR("Unknown connection level")
        if not self.siteSet:
            self.getSiteSEMapping()
        if site not in self.siteSet:
            siteList = [s for s in self.siteSet if ".%s." % site in s]
        else:
            siteList = [site]
        if not siteList:
            return S_ERROR("Unknown site")
        return self._getSEsForSItes(siteList, connectionIndex=connectionIndex)

    def _getSEsForSItes(self, siteList, connectionIndex):
        """Extract list of SEs for a connectivity"""
        mapping = self.getSiteSEMapping()
        if not mapping["OK"]:
            return mapping
        ses = []
        for index in range(LOCAL, connectionIndex + 1):
            for site in siteList:
                ses += mapping["Value"][index].get(site, [])
        if not ses:
            return S_ERROR("No SE found")
        return S_OK(sorted(ses))

    def getSEsAtSite(self, site):
        """Get local SEs"""
        return self.getSEsForSite(site, connectionLevel=LOCAL)

    def isSameSiteSE(self, se1, se2):
        """Are these 2 SEs at the same site"""
        res = self.getLocalSiteForSE(se1)
        if not res["OK"]:
            return res
        site1 = res["Value"]
        res = self.getLocalSiteForSE(se2)
        if not res["OK"]:
            return res
        site2 = res["Value"]
        return S_OK(site1 == site2)

    def getSEsAtCountry(self, country, connectionLevel=None):
        """Get all SEs at a given country"""
        connectionIndex = _getConnectionIndex(connectionLevel, default=DOWNLOAD)
        if connectionIndex is None:
            return S_ERROR("Unknown connection level")
        if not self.siteSet:
            self.getSiteSEMapping()
        siteList = [site for site in self.siteSet if siteCountryName(site) == country.lower()]
        if not siteList:
            return S_ERROR("No SEs found in country")
        return self._getSEsForSItes(siteList, connectionIndex)

    def getSEInGroupAtSite(self, seGroup, site):
        """Get the SE in a group or list of SEs that is present at a site"""
        seList = self.getAllSEsInGroupAtSite(seGroup, site)
        if not seList["OK"] or seList["Value"] is None:
            return seList
        return S_OK(seList["Value"][0])

    def getAllSEsInGroupAtSite(self, seGroup, site):
        """Get all SEs in a group or list of SEs that are present at a site"""
        seList = resolveSEGroup(seGroup)
        if not seList:
            return S_ERROR("SEGroup does not exist")
        sesAtSite = self.getSEsAtSite(site)
        if not sesAtSite["OK"]:
            return sesAtSite
        foundSEs = set(seList) & set(sesAtSite["Value"])
        if not foundSEs:
            gLogger.warn("No SE found at that site", "in group %s at %s" % (seGroup, site))
            return S_OK()
        return S_OK(sorted(foundSEs))

    def getRegistrationProtocols(self):
        """Returns the Favorite registration protocol defined in the CS, or 'srm' as default"""
        return self.__opsHelper.getValue("DataManagement/RegistrationProtocols", ["srm", "dips"])

    def getThirdPartyProtocols(self):
        """Returns the Favorite third party protocol defined in the CS, or 'srm' as default"""
        return self.__opsHelper.getValue("DataManagement/ThirdPartyProtocols", ["srm"])

    def getAccessProtocols(self):
        """Returns the Favorite access protocol defined in the CS, or 'srm' as default"""
        return self.__opsHelper.getValue("DataManagement/AccessProtocols", ["srm", "dips"])

    def getWriteProtocols(self):
        """Returns the Favorite Write protocol defined in the CS, or 'srm' as default"""
        return self.__opsHelper.getValue("DataManagement/WriteProtocols", ["srm", "dips"])

    def getStageProtocols(self):
        """Returns the Favorite staging protocol defined in the CS. There are no default"""
        return self.__opsHelper.getValue("DataManagement/StageProtocols", list())



"""
:mod: DataIntegrityHandler


.. module: DataIntegrityHandler

  :synopsis: DataIntegrityHandler is the implementation of the Data Integrity service in
  the DISET framework

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# imports
import six

# from DIRAC
from DIRAC import S_OK
from DIRAC.Core.DISET.RequestHandler import RequestHandler
from DIRAC.DataManagementSystem.DB.DataIntegrityDB import DataIntegrityDB

__RCSID__ = "$Id$"


class DataIntegrityHandler(RequestHandler):
    """
    .. class:: DataIntegrityHandler

    Implementation of the Data Integrity service in the DISET framework.
    """

    @classmethod
    def initializeHandler(cls, serviceInfoDict):
        """Initialization of DB object"""

        cls.dataIntegrityDB = DataIntegrityDB()
        return S_OK()

    types_removeProblematic = [list(six.integer_types) + [list]]

    def export_removeProblematic(self, fileID):
        """Remove the file with the supplied FileID from the database"""
        if isinstance(fileID, list):
            fileIDs = fileID
        else:
            fileIDs = [int(fileID)]
        self.log.info("DataIntegrityHandler.removeProblematic: Attempting to remove problematic.")
        res = self.dataIntegrityDB.removeProblematic(fileIDs)
        if not res["OK"]:
            self.log.error("DataIntegrityHandler.removeProblematic: Failed to remove problematic.", res["Message"])
        return res

    types_getProblematic = []

    def export_getProblematic(self):
        """Get the next problematic to resolve from the IntegrityDB"""
        self.log.info("DataIntegrityHandler.getProblematic: Getting file to resolve.")
        res = self.dataIntegrityDB.getProblematic()
        if not res["OK"]:
            self.log.error(
                "DataIntegrityHandler.getProblematic: Failed to get problematic file to resolve.", res["Message"]
            )
        return res

    types_getPrognosisProblematics = [list(six.string_types)]

    def export_getPrognosisProblematics(self, prognosis):
        """Get problematic files from the problematics table of the IntegrityDB"""
        self.log.info("DataIntegrityHandler.getPrognosisProblematics: Getting files with %s prognosis." % prognosis)
        res = self.dataIntegrityDB.getPrognosisProblematics(prognosis)
        if not res["OK"]:
            self.log.error(
                "DataIntegrityHandler.getPrognosisProblematics: Failed to get prognosis files.", res["Message"]
            )
        return res

    types_setProblematicStatus = [list(six.integer_types), list(six.string_types)]

    def export_setProblematicStatus(self, fileID, status):
        """Update the status of the problematics with the provided fileID"""
        self.log.info("DataIntegrityHandler.setProblematicStatus: Setting file %s status to %s." % (fileID, status))
        res = self.dataIntegrityDB.setProblematicStatus(fileID, status)
        if not res["OK"]:
            self.log.error("DataIntegrityHandler.setProblematicStatus: Failed to set status.", res["Message"])
        return res

    types_incrementProblematicRetry = [list(six.integer_types)]

    def export_incrementProblematicRetry(self, fileID):
        """Update the retry count for supplied file ID."""
        self.log.info("DataIntegrityHandler.incrementProblematicRetry: Incrementing retries for file %s." % (fileID))
        res = self.dataIntegrityDB.incrementProblematicRetry(fileID)
        if not res["OK"]:
            self.log.error(
                "DataIntegrityHandler.incrementProblematicRetry: Failed to increment retries.", res["Message"]
            )
        return res

    types_insertProblematic = [list(six.string_types), dict]

    def export_insertProblematic(self, source, fileMetadata):
        """Insert problematic files into the problematics table of the IntegrityDB"""
        self.log.info("DataIntegrityHandler.insertProblematic: Inserting problematic file to integrity DB.")
        res = self.dataIntegrityDB.insertProblematic(source, fileMetadata)
        if not res["OK"]:
            self.log.error("DataIntegrityHandler.insertProblematic: Failed to insert.", res["Message"])
        return res

    types_changeProblematicPrognosis = []

    def export_changeProblematicPrognosis(self, fileID, newPrognosis):
        """Change the prognosis for the supplied file"""
        self.log.info("DataIntegrityHandler.changeProblematicPrognosis: Changing problematic prognosis.")
        res = self.dataIntegrityDB.changeProblematicPrognosis(fileID, newPrognosis)
        if not res["OK"]:
            self.log.error("DataIntegrityHandler.changeProblematicPrognosis: Failed to update.", res["Message"])
        return res

    types_getTransformationProblematics = [list(six.integer_types)]

    def export_getTransformationProblematics(self, transID):
        """Get the problematics for a given transformation"""
        self.log.info("DataIntegrityHandler.getTransformationProblematics: Getting problematics for transformation.")
        res = self.dataIntegrityDB.getTransformationProblematics(transID)
        if not res["OK"]:
            self.log.error("DataIntegrityHandler.getTransformationProblematics: Failed.", res["Message"])
        return res

    types_getProblematicsSummary = []

    def export_getProblematicsSummary(self):
        """Get a summary from the Problematics table from the IntegrityDB"""
        self.log.info("DataIntegrityHandler.getProblematicsSummary: Getting problematics summary.")
        res = self.dataIntegrityDB.getProblematicsSummary()
        if res["OK"]:
            for prognosis, statusDict in res["Value"].items():
                self.log.info("DataIntegrityHandler.getProblematicsSummary: %s." % prognosis)
                for status, count in statusDict.items():
                    self.log.info("DataIntegrityHandler.getProblematicsSummary: \t%-10s %-10s." % (status, str(count)))
        else:
            self.log.error("DataIntegrityHandler.getProblematicsSummary: Failed to get summary.", res["Message"])
        return res

    types_getDistinctPrognosis = []

    def export_getDistinctPrognosis(self):
        """Get a list of the distinct prognosis from the IntegrityDB"""
        self.log.info("DataIntegrityHandler.getDistinctPrognosis: Getting distinct prognosis.")
        res = self.dataIntegrityDB.getDistinctPrognosis()
        if res["OK"]:
            for prognosis in res["Value"]:
                self.log.info("DataIntegrityHandler.getDistinctPrognosis: \t%s." % prognosis)
        else:
            self.log.error("DataIntegrityHandler.getDistinctPrognosis: Failed to get unique prognosis.", res["Message"])
        return res



#########################################################################################
# OAR.py
# 10.11.2014
# Author: Matvey Sapunov, A.T.
#########################################################################################

""" OAR.py is a DIRAC independent class representing OAR batch system.
    OAR objects are used as backend batch system representation for
    LocalComputingElement and SSHComputingElement classes
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import subprocess
import shlex
import os
import json

__RCSID__ = "$Id$"


class OAR(object):
    def submitJob(self, **kwargs):
        """Submit nJobs to the OAR batch system"""

        resultDict = {}

        MANDATORY_PARAMETERS = ["Executable", "OutputDir", "ErrorDir", "Queue", "SubmitOptions"]

        for argument in MANDATORY_PARAMETERS:
            if argument not in kwargs:
                resultDict["Status"] = -1
                resultDict["Message"] = "No %s" % argument
                return resultDict

        nJobs = kwargs.get("NJobs")
        if not nJobs:
            nJobs = 1

        outputDir = kwargs["OutputDir"]
        errorDir = kwargs["ErrorDir"]
        queue = kwargs["Queue"]
        submitOptions = kwargs["SubmitOptions"]
        executable = kwargs["Executable"]

        outFile = os.path.join(outputDir, "%jobid%")
        errFile = os.path.join(errorDir, "%jobid%")
        outFile = os.path.expandvars(outFile)
        errFile = os.path.expandvars(errFile)
        executable = os.path.expandvars(executable)
        jobIDs = []
        preamble = kwargs.get("Preamble")
        for _i in range(nJobs):
            cmd = "%s; " % preamble if preamble else ""
            cmd += "oarsub -O %s.out -E %s.err -q %s -n DIRACPilot %s %s" % (
                outFile,
                errFile,
                queue,
                submitOptions,
                executable,
            )
            sp = subprocess.Popen(
                cmd,
                shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                universal_newlines=True,
            )
            output, error = sp.communicate()
            status = sp.returncode

            if status != 0 or not output:
                break

            lines = output.split("\n")
            jid = ""
            if "OAR_JOB_ID" in lines[-1]:
                _prefix, jid = lines[-1].split("=")

            if not jid:
                break

            jid = jid.strip()
            jobIDs.append(jid)

        if jobIDs:
            resultDict["Status"] = 0
            resultDict["Jobs"] = jobIDs
        else:
            resultDict["Status"] = status
            resultDict["Message"] = error
            resultDict["Jobs"] = jobIDs
        return resultDict

    def killJob(self, **kwargs):
        """Delete a job from OAR batch scheduler. Input: list of jobs output: int"""

        resultDict = {}

        MANDATORY_PARAMETERS = ["JobIDList"]
        for argument in MANDATORY_PARAMETERS:
            if argument not in kwargs:
                resultDict["Status"] = -1
                resultDict["Message"] = "No %s" % argument
                return resultDict

        jobIDList = kwargs["JobIDList"]
        if not jobIDList:
            resultDict["Status"] = -1
            resultDict["Message"] = "Empty job list"
            return resultDict

        successful = []
        failed = []
        errors = ""
        for job in jobIDList:
            sp = subprocess.Popen(
                shlex.split("oardel %s" % job),
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                universal_newlines=True,
            )
            output, error = sp.communicate()
            status = sp.returncode
            if status != 0:
                failed.append(job)
                errors += error
            else:
                successful.append(job)

        resultDict["Status"] = 0
        if failed:
            resultDict["Status"] = 1
            resultDict["Message"] = errors
        resultDict["Successful"] = successful
        resultDict["Failed"] = failed
        return resultDict

    def getJobStatus(self, **kwargs):
        """Get status of the jobs in the given list"""

        resultDict = {}

        MANDATORY_PARAMETERS = ["JobIDList"]
        for argument in MANDATORY_PARAMETERS:
            if argument not in kwargs:
                resultDict["Status"] = -1
                resultDict["Message"] = "No %s" % argument
                return resultDict

        jobIDList = kwargs["JobIDList"]
        if not jobIDList:
            resultDict["Status"] = -1
            resultDict["Message"] = "Empty job list"
            return resultDict

        user = kwargs.get("User")
        if not user:
            user = os.environ.get("USER")
        if not user:
            resultDict["Status"] = -1
            resultDict["Message"] = "No user name"
            return resultDict

        cmd = "oarstat --sql \"project = '%s'\" -J" % user
        sp = subprocess.Popen(
            shlex.split(cmd),
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
        )
        output, error = sp.communicate()
        status = sp.returncode
        if status != 0:
            resultDict["Status"] = status
            resultDict["Message"] = error
            return resultDict

        try:
            output = json.loads(output)
        except Exception as x:
            resultDict["Status"] = 2048
            resultDict["Message"] = str(x)
            return resultDict

        if not len(output) > 0:
            resultDict["Status"] = 1024
            resultDict["Message"] = output
            return resultDict

        statusDict = {}
        for job in jobIDList:

            if job not in output:
                statusDict[job] = "Unknown"
                continue

            if "state" not in output[job]:
                statusDict[job] = "Unknown"
                continue
            state = output[job]["state"]

            if state in ["Running", "Finishing"]:
                statusDict[job] = "Running"
                continue

            if state in ["Error", "toError"]:
                statusDict[job] = "Aborted"
                continue

            if state in ["Waiting", "Hold", "toAckReservation", "Suspended", "toLaunch", "Launching"]:
                statusDict[job] = "Waiting"
                continue

            if state == "Terminated":
                statusDict[job] = "Done"
                continue

            statusDict[job] = "Unknown"
            continue

        # Final output
        status = 0
        resultDict["Status"] = 0
        resultDict["Jobs"] = statusDict
        return resultDict

    def getCEStatus(self, **kwargs):
        """Get the overall status of the CE"""

        resultDict = {}

        user = kwargs.get("User")
        if not user:
            user = os.environ.get("USER")
        if not user:
            resultDict["Status"] = -1
            resultDict["Message"] = "No user name"
            return resultDict

        waitingJobs = 0
        runningJobs = 0

        sp = subprocess.Popen(
            shlex.split("oarstat -u %s -J" % user),
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
        )
        output, error = sp.communicate()
        status = sp.returncode
        if status != 0:
            if "arrayref expected" in output:
                resultDict["Status"] = 0
                resultDict["Waiting"] = 0
                resultDict["Running"] = 0
                return resultDict
            resultDict["Status"] = status
            resultDict["Message"] = error
            return resultDict

        try:
            output = json.loads(output)
        except Exception as x:
            resultDict["Status"] = 2048
            resultDict["Message"] = str(x)
            return resultDict

        if output > 0:
            resultDict["Status"] = 0
            resultDict["Waiting"] = waitingJobs
            resultDict["Running"] = runningJobs
            return resultDict

        for value in output.values():

            if "state" not in value:
                continue
            state = value["state"]

            if state in ["Running", "Finishing"]:
                runningJobs += 1
                continue

            if state in ["Waiting", "Hold", "toAckReservation", "Suspended", "toLaunch", "Launching"]:
                waitingJobs += 1
                continue

        # Final output
        resultDict["Status"] = 0
        resultDict["Waiting"] = waitingJobs
        resultDict["Running"] = runningJobs
        return resultDict



#!/usr/bin/env python
########################################################################
# File :    dirac-admin-modify-user
# Author :  Adrian Casajus
########################################################################
"""
Modify a user in the CS.

Example:
  $ dirac-admin-modify-user vhamar /C=FR/O=Org/CN=User dirac_user
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    Script.registerSwitch("p:", "property=", "Add property to the user <name>=<value>")
    Script.registerSwitch("f", "force", "create the user if it doesn't exist")
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(" user:     User name")
    Script.registerArgument(" DN:       DN of the User")
    Script.registerArgument(["group:    Add the user to the group"])
    Script.parseCommandLine(ignoreErrors=True)

    from DIRAC.Interfaces.API.DiracAdmin import DiracAdmin

    diracAdmin = DiracAdmin()
    exitCode = 0
    forceCreation = False
    errorList = []

    userProps = {}
    for unprocSw in Script.getUnprocessedSwitches():
        if unprocSw[0] in ("f", "force"):
            forceCreation = True
        elif unprocSw[0] in ("p", "property"):
            prop = unprocSw[1]
            pl = prop.split("=")
            if len(pl) < 2:
                errorList.append(
                    ("in arguments", "Property %s has to include a '=' to separate name from value" % prop)
                )
                exitCode = 255
            else:
                pName = pl[0]
                pValue = "=".join(pl[1:])
                print("Setting property %s to %s" % (pName, pValue))
                userProps[pName] = pValue

    userName, userProps["DN"], userProps["Groups"] = Script.getPositionalArgs(group=True)

    if not diracAdmin.csModifyUser(userName, userProps, createIfNonExistant=forceCreation):
        errorList.append(("modify user", "Cannot modify user %s" % userName))
        exitCode = 255
    else:
        result = diracAdmin.csCommitChanges()
        if not result["OK"]:
            errorList.append(("commit", result["Message"]))
            exitCode = 255

    for error in errorList:
        print("ERROR %s: %s" % error)

    DIRAC.exit(exitCode)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""
Select/Add/Delete a new DownTime entry for a given Site or Service.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import datetime

from DIRAC import gLogger, exit as DIRACExit, version
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script
from DIRAC.Core.Utilities import Time
from DIRAC.Core.Utilities.PrettyPrint import printTable
from DIRAC.ResourceStatusSystem.Utilities import Utils


subLogger = None


def registerSwitches():
    """
    Registers all switches that can be used while calling the script from the
    command line interface.
    """

    switches = (
        ("downtimeID=", "ID of the downtime"),
        ("element=", "Element (Site, Service) affected by the downtime"),
        ("name=", "Name of the element"),
        ("startDate=", "Starting date of the downtime"),
        ("endDate=", "Ending date of the downtime"),
        ("severity=", "Severity of the downtime (Warning, Outage)"),
        ("description=", "Description of the downtime"),
        ("link=", "URL of the downtime announcement"),
        ("ongoing", 'To force "select" to return the ongoing downtimes'),
    )

    for switch in switches:
        Script.registerSwitch("", switch[0], switch[1])


def registerUsageMessage():
    """
    Takes the script __doc__ and adds the DIRAC version to it
    """

    usageMessage = "DIRAC version: %s \n" % version
    usageMessage += __doc__

    Script.setUsageMessage(usageMessage)


def parseSwitches():
    """
    Parses the arguments passed by the user
    """

    Script.parseCommandLine(ignoreErrors=True)
    args = Script.getPositionalArgs()
    if not args:
        error("Missing mandatory 'query' argument")
    elif not args[0].lower() in ("select", "add", "delete"):
        error("Missing mandatory argument")
    else:
        query = args[0].lower()

    switches = dict(Script.getUnprocessedSwitches())

    # Default values
    switches.setdefault("downtimeID", None)
    switches.setdefault("element", None)
    switches.setdefault("name", None)
    switches.setdefault("startDate", None)
    switches.setdefault("endDate", None)
    switches.setdefault("severity", None)
    switches.setdefault("description", None)
    switches.setdefault("link", None)

    if query in ("add", "delete") and switches["downtimeID"] is None:
        error("'downtimeID' switch is mandatory for '%s' but found missing" % query)

    if query in ("add", "delete") and "ongoing" in switches:
        error("'ongoing' switch can be used only with 'select'")

    subLogger.debug("The switches used are:")
    map(subLogger.debug, switches.items())

    return (args, switches)


# UTILS: for filtering 'select' output


def filterDate(selectOutput, start, end):
    """
    Selects all the downtimes that meet the constraints of 'start' and 'end' dates
    """

    downtimes = selectOutput
    downtimesFiltered = []

    if start is not None:
        try:
            start = Time.fromString(start)
        except Exception:
            error("datetime formt is incorrect, pls try [%Y-%m-%d[ %H:%M:%S]]")
        start = Time.toEpoch(start)

    if end is not None:
        try:
            end = Time.fromString(end)
        except Exception:
            error("datetime formt is incorrect, pls try [%Y-%m-%d[ %H:%M:%S]]")
        end = Time.toEpoch(end)

    if start is not None and end is not None:
        for dt in downtimes:
            dtStart = Time.toEpoch(dt["startDate"])
            dtEnd = Time.toEpoch(dt["endDate"])
            if (dtStart >= start) and (dtEnd <= end):
                downtimesFiltered.append(dt)

    elif start is not None and end is None:
        for dt in downtimes:
            dtStart = Time.toEpoch(dt["startDate"])
            if dtStart >= start:
                downtimesFiltered.append(dt)

    elif start is None and end is not None:
        for dt in downtimes:
            dtEnd = Time.toEpoch(dt["endDate"])
            if dtEnd <= end:
                downtimesFiltered.append(dt)

    else:
        downtimesFiltered = downtimes

    return downtimesFiltered


def filterOngoing(selectOutput):
    """
    Selects all the ongoing downtimes
    """

    downtimes = selectOutput
    downtimesFiltered = []
    currentDate = Time.toEpoch(Time.dateTime())

    for dt in downtimes:
        dtStart = Time.toEpoch(dt["startDate"])
        dtEnd = Time.toEpoch(dt["endDate"])
        if (dtStart <= currentDate) and (dtEnd >= currentDate):
            downtimesFiltered.append(dt)

    return downtimesFiltered


def filterDescription(selectOutput, description):
    """
    Selects all the downtimes that match 'description'
    """

    downtimes = selectOutput
    downtimesFiltered = []
    if description is not None:
        for dt in downtimes:
            if description in dt["description"]:
                downtimesFiltered.append(dt)
    else:
        downtimesFiltered = downtimes

    return downtimesFiltered


# Utils: for formatting query output and notifications


def error(msg):
    """
    Format error messages
    """

    subLogger.error("\nERROR:")
    subLogger.error("\t" + msg)
    subLogger.error("\tPlease, check documentation below")
    Script.showHelp(exitCode=1)


def confirm(query, matches):
    """
    Format confirmation messages
    """

    subLogger.notice("\nNOTICE: '%s' request successfully executed ( matches' number: %s )! \n" % (query, matches))


def tabularPrint(table):

    columns_names = list(table[0])
    records = []
    for row in table:
        record = []
        for k, v in row.items():
            if isinstance(v, datetime.datetime):
                record.append(Time.toString(v))
            elif v is None:
                record.append("")
            else:
                record.append(v)
        records.append(record)

    output = printTable(columns_names, records, numbering=False, columnSeparator=" | ", printOut=False)

    subLogger.notice(output)


def select(switchDict):
    """
    Given the switches, request a query 'select' on the ResourceManagementDB
    that gets from DowntimeCache all rows that match the parameters given.
    """

    rmsClient = ResourceManagementClient()

    meta = {
        "columns": [
            "DowntimeID",
            "Element",
            "Name",
            "StartDate",
            "EndDate",
            "Severity",
            "Description",
            "Link",
            "DateEffective",
        ]
    }

    result = {"output": None, "OK": None, "Message": None, "match": None}
    output = rmsClient.selectDowntimeCache(
        downtimeID=switchDict["downtimeID"],
        element=switchDict["element"],
        name=switchDict["name"],
        severity=switchDict["severity"],
        meta=meta,
    )

    if not output["OK"]:
        return output
    result["output"] = [dict(zip(output["Columns"], dt)) for dt in output["Value"]]
    if "ongoing" in switchDict:
        result["output"] = filterOngoing(result["output"])
    else:
        result["output"] = filterDate(result["output"], switchDict["startDate"], switchDict["endDate"])
    result["output"] = filterDescription(result["output"], switchDict["description"])
    result["match"] = len(result["output"])
    result["OK"] = True
    result["message"] = output["Message"] if "Message" in output else None

    return result


def add(switchDict):
    """
    Given the switches, request a query 'addOrModify' on the ResourceManagementDB
    that inserts or updates-if-duplicated from DowntimeCache.
    """

    rmsClient = ResourceManagementClient()

    result = {"output": None, "OK": None, "Message": None, "match": None}
    output = rmsClient.addOrModifyDowntimeCache(
        downtimeID=switchDict["downtimeID"],
        element=switchDict["element"],
        name=switchDict["name"],
        startDate=switchDict["startDate"],
        endDate=switchDict["endDate"],
        severity=switchDict["severity"],
        description=switchDict["description"],
        link=switchDict["link"],
    )

    if not output["OK"]:
        return output

    if output["Value"]:
        result["match"] = int(output["Value"])
    result["OK"] = True
    result["message"] = output["Message"] if "Message" in output else None

    return result


def delete(switchDict):
    """
    Given the switches, request a query 'delete' on the ResourceManagementDB
    that deletes from DowntimeCache all rows that match the parameters given.
    """

    rmsClient = ResourceManagementClient()

    result = {"output": None, "OK": None, "Message": None, "match": None}
    output = rmsClient.deleteDowntimeCache(
        downtimeID=switchDict["downtimeID"],
        element=switchDict["element"],
        name=switchDict["name"],
        startDate=switchDict["startDate"],
        endDate=switchDict["endDate"],
        severity=switchDict["severity"],
        description=switchDict["description"],
        link=switchDict["link"],
    )
    if not output["OK"]:
        return output

    if output["Value"]:
        result["match"] = int(output["Value"])
    result["OK"] = True
    result["Message"] = output["Message"] if "Message" in output else None

    return result


def run(args, switchDict):
    """
    Main function of the script
    """

    query = args[0]

    # it exectues the query request: e.g. if it's a 'select' it executes 'select()'
    # the same if it is add, delete
    result = eval(query + "( switchDict )")

    if result["OK"]:
        if query == "select" and result["match"] > 0:
            tabularPrint(result["output"])
        confirm(query, result["match"])
    else:
        error(result["Message"])


@Script()
def main():
    global subLogger
    global ResourceManagementClient

    subLogger = gLogger.getSubLogger(__file__)

    # Script initialization
    registerSwitches()
    registerUsageMessage()
    args, switchDict = parseSwitches()

    ResourceManagementClient = getattr(
        Utils.voimport("DIRAC.ResourceStatusSystem.Client.ResourceManagementClient"),
        "ResourceManagementClient",
    )

    # Run script
    run(args, switchDict)

    # Bye
    DIRACExit(0)


if __name__ == "__main__":
    main()



#!/bin/env python
"""
tests for InProcessComputingElement module
"""

import os
import shutil

import pytest

from DIRAC.Resources.Computing.test.Test_PoolComputingElement import jobScript, _stopJob
from DIRAC.WorkloadManagementSystem.Utilities.Utils import createJobWrapper

# sut
from DIRAC.Resources.Computing.InProcessComputingElement import InProcessComputingElement


@pytest.mark.slow
def test_submitJob():
    with open("testJob.py", "w") as execFile:
        execFile.write(jobScript % "1")
    os.chmod("testJob.py", 0o755)

    ce = InProcessComputingElement("InProcessCE")
    res = ce.submitJob("testJob.py", None)
    assert res["OK"] is True
    res = ce.getCEStatus()
    assert res["OK"] is True
    assert res["SubmittedJobs"] == 1
    _stopJob(1)
    for ff in ["testJob.py", "stop_job_2", "job.info", "std.out"]:
        if os.path.isfile(ff):
            os.remove(ff)

    #
    # With a job wrapper and some MP parameters
    with open("testJob.py", "w") as execFile:
        execFile.write(jobScript % "2")
    os.chmod("testJob.py", 0o755)

    jobParams = {"JobType": "User", "Executable": "testJob.py"}
    resourceParams = {"GridCE": "some_CE"}
    optimizerParams = {}

    wrapperFile = createJobWrapper(2, jobParams, resourceParams, optimizerParams, logLevel="DEBUG")[
        "Value"
    ]  # This is not under test, assuming it works fine
    res = ce.submitJob(
        wrapperFile,
        proxy=None,
        numberOfProcessors=4,
        maxNumberOfProcessors=8,
        wholeNode=False,
        mpTag=True,
        jobDesc={"jobParams": jobParams, "resourceParams": resourceParams, "optimizerParams": optimizerParams},
    )
    assert res["OK"] is True

    res = ce.getCEStatus()
    assert res["OK"] is True
    assert res["SubmittedJobs"] == 2

    _stopJob(2)
    for ff in ["testJob.py", "stop_job_2", "job.info", "std.out"]:
        if os.path.isfile(ff):
            os.remove(ff)
    if os.path.isdir("job"):
        shutil.rmtree("job")



""" The SystemAdministratorIntegrator is a class integrating access to all the
    SystemAdministrator services configured in the system
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC import S_OK
from DIRAC.Core.Utilities.ThreadPool import ThreadPool
from DIRAC.FrameworkSystem.Client.SystemAdministratorClient import SystemAdministratorClient
import DIRAC.ConfigurationSystem.Client.Helpers.Registry as Registry

SYSADMIN_PORT = 9162


class SystemAdministratorIntegrator(object):
    def __init__(self, **kwargs):
        """Constructor"""
        if "hosts" in kwargs:
            self.__hosts = kwargs["hosts"]
            del kwargs["hosts"]
        else:
            result = Registry.getHosts()
            if result["OK"]:
                self.__hosts = result["Value"]
            else:
                self.__hosts = []
            # Excluded hosts
            if "exclude" in kwargs:
                self.__hosts = list(set(self.__hosts) - set(kwargs["exclude"]))

        # Ping the hosts to remove those that don't have a SystemAdministrator service
        sysAdminHosts = []
        self.silentHosts = []
        self.__resultDict = {}
        self.__kwargs = {}
        pool = ThreadPool(len(self.__hosts))
        for host in self.__hosts:
            pool.generateJobAndQueueIt(
                self.__executeClient, args=[host, "ping"], kwargs={}, oCallback=self.__processResult
            )

        pool.processAllResults()
        for host, result in self.__resultDict.items():
            if result["OK"]:
                sysAdminHosts.append(host)
            else:
                self.silentHosts.append(host)
        del pool

        self.__hosts = sysAdminHosts

        self.__kwargs = dict(kwargs)
        self.__pool = ThreadPool(len(self.__hosts))
        self.__resultDict = {}

    def getSilentHosts(self):
        """Get a list of non-responding hosts

        :return: list of hosts
        """
        return self.silentHosts

    def getRespondingHosts(self):
        """Get a list of responding hosts

        :return: list of hosts
        """
        return self.__hosts

    def __getattr__(self, name):
        self.call = name
        return self.execute

    def __executeClient(self, host, method, *parms, **kwargs):
        """Execute RPC method on a given host"""
        hostName = Registry.getHostOption(host, "Host", host)
        client = SystemAdministratorClient(hostName, **self.__kwargs)
        result = getattr(client, method)(*parms, **kwargs)
        result["Host"] = host
        return result

    def __processResult(self, id_, result):
        """Collect results in the final structure"""
        host = result["Host"]
        del result["Host"]
        self.__resultDict[host] = result

    def execute(self, *args, **kwargs):
        """Main execution method"""
        self.__resultDict = {}
        for host in self.__hosts:
            self.__pool.generateJobAndQueueIt(
                self.__executeClient, args=[host, self.call] + list(args), kwargs=kwargs, oCallback=self.__processResult
            )

        self.__pool.processAllResults()
        return S_OK(self.__resultDict)



""" Test class for SGEResourceUsage utility
"""

import pytest

from DIRAC import S_OK

# sut
from DIRAC.Resources.Computing.BatchSystems.TimeLeft.SGEResourceUsage import SGEResourceUsage

RESULT_FROM_SGE = """
job_number:                 7448711
exec_file:                  job_scripts/7448711
submission_time:            Fri Feb 26 19:56:58 2021
owner:                      pltlhcb001
uid:                        110476
group:                      pltlhcb
gid:                        110013
sge_o_path:                 /usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin
sge_o_host:                 grendel
account:                    sge
merge:                      y
hard resource_list:         decom=FALSE,s_vmem=3994M,h_vmem=3994M
mail_list:                  pltlhcb001@grendel.private.dns.zone
notify:                     FALSE
job_name:                   arc60158374
priority:                   -512
jobshare:                   0
hard_queue_list:            grid7
restart:                    n
shell_list:                 NONE:/bin/sh
env_list:                   TERM=NONE
script_file:                STDIN
parallel environment:  smp range: 8
project:                    grid
binding:                    NONE
job_type:                   NONE
usage         1:            cpu=00:00:23, mem=0.76117 GB s, io=0.08399 GB, vmem=384.875M, maxvmem=384.875M
binding       1:            NONE
scheduling info:            (Collecting of scheduler job information is turned off)
"""


@pytest.mark.parametrize(
    "runCommandResult, cpuLimitExpected, wallClockLimitExpected",
    [((S_OK(RESULT_FROM_SGE), S_OK("bof")), (720 * 60 / 2.5), (1440 * 60 / 2.5))],
)
def test_init(mocker, runCommandResult, cpuLimitExpected, wallClockLimitExpected):
    mocker.patch(
        "DIRAC.Resources.Computing.BatchSystems.TimeLeft.SGEResourceUsage.runCommand", side_effect=runCommandResult
    )

    sgeResourceUsage = SGEResourceUsage()
    res = sgeResourceUsage.getResourceUsage()
    assert not res["OK"], res["Message"]



"""
ColoredBaseFormatter
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import sys

from DIRAC.FrameworkSystem.private.standardLogging.Formatter.BaseFormatter import BaseFormatter


class ColoredBaseFormatter(BaseFormatter):
    """
    ColoredBaseFormatter is used to format log record to create a string representing a log message.
    It is based on the BaseFormatter object which is based on the of the standard logging library.

    This custom formatter is useful for format messages to correspond with the gLogger format.
    It adds color on all messages which come from StdoutBackend and StderrBackend
    and color them according to their levels.
    """

    COLOR_MAP = {"black": 0, "red": 1, "green": 2, "yellow": 3, "blue": 4, "magenta": 5, "cyan": 6, "white": 7}

    LEVEL_MAP = {
        "ALWAYS": ("black", "white", False),
        "NOTICE": (None, "magenta", False),
        "INFO": (None, "green", False),
        "VERBOSE": (None, "cyan", False),
        "DEBUG": (None, "blue", False),
        "WARN": (None, "yellow", False),
        "ERROR": (None, "red", False),
        "FATAL": ("red", "black", False),
    }

    def __init__(self, fmt=None, datefmt=None):
        """
        Initialize the formatter without using parameters.
        They are then modified in format()

        :param str fmt: log format: "%(asctime)s UTC %(name)s %(levelname)s: %(message)"
        :param str datefmt: date format: "%Y-%m-%d %H:%M:%S"
        """
        super(ColoredBaseFormatter, self).__init__()

    def format(self, record):
        """
        Overriding.
        format is the main method of the Formatter object because it is the method which transforms
        a log record into a string with colors.
        According to the level, the method get colors from LEVEL_MAP to add them to the message.

        :param record: the log record containing all the information about the log message: name, level, threadid...
        """

        stringRecord = super(ColoredBaseFormatter, self).format(record)

        # post treatment
        if record.color and sys.stdout.isatty() and sys.stderr.isatty():
            params = []
            bg, fg, bold = self.LEVEL_MAP[record.levelname]
            if bg in self.COLOR_MAP:
                params.append(str(self.COLOR_MAP[bg] + 40))
            if fg in self.COLOR_MAP:
                params.append(str(self.COLOR_MAP[fg] + 30))
            if bold:
                params.append("1")
            stringRecord = "".join(("\x1b[", ";".join(params), "m", stringRecord, "\x1b[0m"))

        return stringRecord
